import random
from typing import Any

from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_list import (
    FunctionArgumentList,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.statesfunction.states_function import (
    StatesFunction,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.state_function_name_types import (
    StatesFunctionNameType,
)
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.states_function_name import (
    StatesFunctionName,
)
from moto.stepfunctions.parser.asl.eval.environment import Environment


class MathRandom(StatesFunction):
    # Returns a random number between the specified start and end number.
    #
    # For example:
    # With input
    # {
    #    "start": 1,
    #    "end": 999
    # }
    #
    # Call
    # "random.$": "States.MathRandom($.start, $.end)"
    #
    # Returns
    # {"random": 456 }

    def __init__(self, arg_list: FunctionArgumentList):
        super().__init__(
            states_name=StatesFunctionName(
                function_type=StatesFunctionNameType.MathRandom
            ),
            arg_list=arg_list,
        )
        if arg_list.size < 2 or arg_list.size > 3:
            raise ValueError(
                f"Expected 2-3 arguments for function type '{type(self)}', but got: '{arg_list}'."
            )

    @staticmethod
    def _validate_integer_value(value: Any, argument_name: str) -> int:
        if not isinstance(value, (int, float)):
            raise TypeError(
                f"Expected integer value for {argument_name}, but got: '{value}'."
            )
        # If you specify a non-integer value for the start number or end number argument,
        # Step Functions will round it off to the nearest integer.
        return int(value)

    def _eval_body(self, env: Environment) -> None:
        self.arg_list.eval(env=env)
        args = env.stack.pop()

        seed = None
        if self.arg_list.size == 3:
            seed = args.pop()
            self._validate_integer_value(seed, "seed")

        end = self._validate_integer_value(args.pop(), "end")
        start = self._validate_integer_value(args.pop(), "start")

        rand_gen = random.Random(seed)
        rand_int = rand_gen.randint(start, end)
        env.stack.append(rand_int)



from typing import Any

from moto.stepfunctions.parser.asl.component.state.exec.state_map.item_reader.reader_config.max_items_decl import (
    MaxItemsDecl,
)
from moto.stepfunctions.parser.asl.parse.typed_props import TypedProps


class ReaderConfigProps(TypedProps):
    _UNIQUE_SUB_INSTANCES = {MaxItemsDecl}
    name: str

    def add(self, instance: Any) -> None:
        inst_type = type(instance)

        # Subclasses
        for typ in self._UNIQUE_SUB_INSTANCES:
            if issubclass(inst_type, typ):
                super()._add(typ, instance)
                return

        # Base and delegate to preprocessor.
        super().add(instance)



import json

import boto3

from moto import mock_aws


@mock_aws
def test_tagging():
    logs_client = boto3.client("logs", region_name="us-east-1")
    cf_client = boto3.client("cloudformation", region_name="us-east-1")

    template = {
        "AWSTemplateFormatVersion": "2010-09-09",
        "Resources": {
            "testGroup": {
                "Type": "AWS::Logs::LogGroup",
                "Properties": {"Tags": [{"Key": "foo", "Value": "bar"}]},
            }
        },
    }
    template_json = json.dumps(template)
    cf_client.create_stack(
        StackName="test_stack",
        TemplateBody=template_json,
    )

    arn = logs_client.describe_log_groups()["logGroups"][0]["arn"]
    tags = logs_client.list_tags_for_resource(resourceArn=arn)["tags"]
    assert tags == {"foo": "bar"}

    cf_client.delete_stack(StackName="test_stack")
    assert logs_client.describe_log_groups()["logGroups"] == []



from typing import TYPE_CHECKING, List

if TYPE_CHECKING:
    from ..models import FakeSecret


def name_filter(secret: "FakeSecret", names: List[str]) -> bool:
    return _matcher(names, [secret.name])


def description_filter(secret: "FakeSecret", descriptions: List[str]) -> bool:
    return _matcher(descriptions, [secret.description], match_prefix=False)  # type: ignore


def tag_key(secret: "FakeSecret", tag_keys: List[str]) -> bool:
    if not secret.tags:
        return False
    return _matcher(tag_keys, [tag["Key"] for tag in secret.tags])


def tag_value(secret: "FakeSecret", tag_values: List[str]) -> bool:
    if not secret.tags:
        return False
    return _matcher(tag_values, [tag["Value"] for tag in secret.tags])


def filter_all(secret: "FakeSecret", values: List[str]) -> bool:
    attributes = [secret.name, secret.description]
    if secret.tags:
        attributes += [tag["Key"] for tag in secret.tags] + [
            tag["Value"] for tag in secret.tags
        ]

    return _matcher(values, attributes)  # type: ignore


def _matcher(
    patterns: List[str], strings: List[str], match_prefix: bool = True
) -> bool:
    for pattern in [p for p in patterns if p.startswith("!")]:
        for string in strings:
            if not _match_pattern(pattern[1:], string, match_prefix):
                return True

    for pattern in [p for p in patterns if not p.startswith("!")]:
        for string in strings:
            if _match_pattern(pattern, string, match_prefix):
                return True
    return False


def _match_pattern(pattern: str, value: str, match_prefix: bool = True) -> bool:
    if match_prefix:
        return value.startswith(pattern)
    else:
        pattern_words = pattern.split(" ")
        value_words = value.split(" ")
        for pattern_word in pattern_words:
            # all words in value must start with pattern_word
            if not any(
                value_word.startswith(pattern_word) for value_word in value_words
            ):
                return False
    return True



from __future__ import annotations

import abc
from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component


class ErrorName(Component, abc.ABC):
    def __init__(self, error_name: str):
        self.error_name: Final[str] = error_name

    def matches(self, error_name: str) -> bool:
        return self.error_name == error_name

    def __eq__(self, other):
        if isinstance(other, ErrorName):
            return self.matches(other.error_name)
        return False



from moto.cloudformation.parsing import get_references_from_template


class TestParametersFromResource:
    def test_no_properties(self):
        template = {"Type": "AWS::S3::Bucket"}
        assert get_references_from_template(template) == []

    def test_properties_without_param(self):
        template = {"Type": "AWS::S3::Bucket", "Properties": {"BucketName": "mybucket"}}
        assert get_references_from_template(template) == []

    def test_properties_with_param(self):
        template = {
            "Type": "AWS::S3::Bucket",
            "Properties": {
                "BucketName": {"Ref": "BucketName"},
                "Tags": [{"Key": {"Ref": "KeyName"}, "Value": {"Ref": "KeyDesc"}}],
            },
        }
        assert get_references_from_template(template) == [
            "BucketName",
            "KeyName",
            "KeyDesc",
        ]

    def test_nested_properties_with_param(self):
        template = {
            "Type": "AWS::S3::Bucket",
            "Properties": {"L1": {"L2": {"Ref": "Name"}}},
        }
        assert get_references_from_template(template) == ["Name"]



import csv
import datetime
from unittest import SkipTest

import boto3
from dateutil.parser import parse

from moto import mock_aws, settings
from moto.iam.models import IAMBackend, iam_backends
from tests import DEFAULT_ACCOUNT_ID


@mock_aws
def test_invoking_ec2_mark_access_key_as_used():
    c_iam = boto3.client("iam", region_name="us-east-1")
    c_iam.create_user(Path="my/path", UserName="fakeUser")
    key = c_iam.create_access_key(UserName="fakeUser")

    c_ec2 = boto3.client(
        "ec2",
        region_name="us-east-2",
        aws_access_key_id=key["AccessKey"]["AccessKeyId"],
        aws_secret_access_key=key["AccessKey"]["SecretAccessKey"],
    )
    c_ec2.describe_instances()

    last_used = c_iam.get_access_key_last_used(
        AccessKeyId=key["AccessKey"]["AccessKeyId"]
    )["AccessKeyLastUsed"]
    assert "LastUsedDate" in last_used
    assert last_used["ServiceName"] == "ec2"
    assert last_used["Region"] == "us-east-2"


@mock_aws
def test_mark_role_as_last_used():
    role_name = "role_name_created_jan_1st"
    iam = boto3.client("iam", "us-east-1")
    sts = boto3.client("sts", "us-east-1")

    role_arn = iam.create_role(RoleName=role_name, AssumeRolePolicyDocument="example")[
        "Role"
    ]["Arn"]

    creds = sts.assume_role(RoleArn=role_arn, RoleSessionName="temp_session")[
        "Credentials"
    ]

    iam2 = boto3.client(
        "iam",
        "us-east-1",
        aws_access_key_id=creds["AccessKeyId"],
        aws_secret_access_key=creds["SecretAccessKey"],
        aws_session_token=creds["SessionToken"],
    )

    iam2.create_role(RoleName="name", AssumeRolePolicyDocument="example")

    role = iam.get_role(RoleName=role_name)["Role"]
    assert isinstance(role["RoleLastUsed"]["LastUsedDate"], datetime.datetime)

    if not settings.TEST_SERVER_MODE:
        iam: IAMBackend = iam_backends[DEFAULT_ACCOUNT_ID]["global"]
        assert iam.get_role(role_name).last_used is not None


@mock_aws
def test_get_credential_report_content__set_last_used_automatically():
    if not settings.TEST_DECORATOR_MODE:
        raise SkipTest("No point testing this in ServerMode")
    # Ensure LAST_USED field is set
    c_iam = boto3.client("iam", region_name="us-east-1")
    c_iam.create_user(Path="my/path", UserName="fakeUser")
    key = c_iam.create_access_key(UserName="fakeUser")

    c_ec2 = boto3.client(
        "ec2",
        region_name="us-east-2",
        aws_access_key_id=key["AccessKey"]["AccessKeyId"],
        aws_secret_access_key=key["AccessKey"]["SecretAccessKey"],
    )
    c_ec2.describe_instances()

    # VERIFY last_used can be retrieved
    conn = boto3.client("iam", region_name="us-east-1")

    result = conn.generate_credential_report()
    while result["State"] != "COMPLETE":
        result = conn.generate_credential_report()
    result = conn.get_credential_report()
    report = result["Content"].decode("utf-8")
    report_dict = csv.DictReader(report.split("\n"))
    user = next(report_dict)

    assert parse(user["access_key_1_last_used_date"])



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


# RegisterDomain endpoint
@mock_aws
def test_register_domain_boto3():
    client = boto3.client("swf", region_name="us-west-1")
    client.register_domain(
        name="test-domain",
        workflowExecutionRetentionPeriodInDays="60",
        description="A test domain",
    )

    all_domains = client.list_domains(registrationStatus="REGISTERED")
    assert len(all_domains["domainInfos"]) == 1
    domain = all_domains["domainInfos"][0]

    assert domain["name"] == "test-domain"
    assert domain["status"] == "REGISTERED"
    assert domain["description"] == "A test domain"
    assert domain["arn"] == f"arn:aws:swf:us-west-1:{ACCOUNT_ID}:/domain/test-domain"


@mock_aws
def test_register_already_existing_domain_boto3():
    client = boto3.client("swf", region_name="us-west-1")
    client.register_domain(
        name="test-domain",
        workflowExecutionRetentionPeriodInDays="60",
        description="A test domain",
    )

    with pytest.raises(ClientError) as ex:
        client.register_domain(
            name="test-domain",
            workflowExecutionRetentionPeriodInDays="60",
            description="A test domain",
        )
    assert ex.value.response["Error"]["Code"] == "DomainAlreadyExistsFault"
    assert ex.value.response["Error"]["Message"] == "test-domain"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400


# ListDomains endpoint
@mock_aws
def test_list_domains_order_boto3():
    client = boto3.client("swf", region_name="us-west-1")
    client.register_domain(
        name="b-test-domain", workflowExecutionRetentionPeriodInDays="60"
    )
    client.register_domain(
        name="a-test-domain", workflowExecutionRetentionPeriodInDays="60"
    )
    client.register_domain(
        name="c-test-domain", workflowExecutionRetentionPeriodInDays="60"
    )

    all_domains = client.list_domains(registrationStatus="REGISTERED")
    assert len(all_domains["domainInfos"]) == 3

    names = [domain["name"] for domain in all_domains["domainInfos"]]
    assert names == ["a-test-domain", "b-test-domain", "c-test-domain"]


@mock_aws
def test_list_domains_reverse_order_boto3():
    client = boto3.client("swf", region_name="us-west-1")
    client.register_domain(
        name="b-test-domain", workflowExecutionRetentionPeriodInDays="60"
    )
    client.register_domain(
        name="a-test-domain", workflowExecutionRetentionPeriodInDays="60"
    )
    client.register_domain(
        name="c-test-domain", workflowExecutionRetentionPeriodInDays="60"
    )

    all_domains = client.list_domains(
        registrationStatus="REGISTERED", reverseOrder=True
    )
    assert len(all_domains["domainInfos"]) == 3

    names = [domain["name"] for domain in all_domains["domainInfos"]]
    assert names == ["c-test-domain", "b-test-domain", "a-test-domain"]


# DeprecateDomain endpoint
@mock_aws
def test_deprecate_domain_boto3():
    client = boto3.client("swf", region_name="us-west-1")
    client.register_domain(
        name="test-domain", workflowExecutionRetentionPeriodInDays="60"
    )
    client.deprecate_domain(name="test-domain")

    all_domains = client.list_domains(registrationStatus="REGISTERED")
    assert len(all_domains["domainInfos"]) == 0

    all_domains = client.list_domains(registrationStatus="DEPRECATED")
    assert len(all_domains["domainInfos"]) == 1

    domain = all_domains["domainInfos"][0]
    assert domain["name"] == "test-domain"


@mock_aws
def test_deprecate_already_deprecated_domain_boto3():
    client = boto3.client("swf", region_name="us-west-1")
    client.register_domain(
        name="test-domain", workflowExecutionRetentionPeriodInDays="60"
    )
    client.deprecate_domain(name="test-domain")

    with pytest.raises(ClientError) as ex:
        client.deprecate_domain(name="test-domain")
    assert ex.value.response["Error"]["Code"] == "DomainDeprecatedFault"
    assert ex.value.response["Error"]["Message"] == "test-domain"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400


@mock_aws
def test_deprecate_non_existent_domain_boto3():
    client = boto3.client("swf", region_name="us-west-1")

    with pytest.raises(ClientError) as ex:
        client.deprecate_domain(name="non-existent")
    assert ex.value.response["Error"]["Code"] == "UnknownResourceFault"
    assert ex.value.response["Error"]["Message"] == "Unknown domain: non-existent"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400


# UndeprecateDomain endpoint
@mock_aws
def test_undeprecate_domain():
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(
        name="test-domain", workflowExecutionRetentionPeriodInDays="60"
    )
    client.deprecate_domain(name="test-domain")
    client.undeprecate_domain(name="test-domain")

    resp = client.describe_domain(name="test-domain")

    assert resp["domainInfo"]["status"] == "REGISTERED"


@mock_aws
def test_undeprecate_already_undeprecated_domain():
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(
        name="test-domain", workflowExecutionRetentionPeriodInDays="60"
    )
    client.deprecate_domain(name="test-domain")
    client.undeprecate_domain(name="test-domain")

    with pytest.raises(ClientError):
        client.undeprecate_domain(name="test-domain")


@mock_aws
def test_undeprecate_never_deprecated_domain():
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(
        name="test-domain", workflowExecutionRetentionPeriodInDays="60"
    )

    with pytest.raises(ClientError):
        client.undeprecate_domain(name="test-domain")


@mock_aws
def test_undeprecate_non_existent_domain():
    client = boto3.client("swf", region_name="us-east-1")

    with pytest.raises(ClientError):
        client.undeprecate_domain(name="non-existent")


# DescribeDomain endpoint
@mock_aws
def test_describe_domain_boto3():
    client = boto3.client("swf", region_name="us-east-1")
    client.register_domain(
        name="test-domain",
        workflowExecutionRetentionPeriodInDays="60",
        description="A test domain",
    )

    domain = client.describe_domain(name="test-domain")
    assert domain["configuration"]["workflowExecutionRetentionPeriodInDays"] == "60"
    assert domain["domainInfo"]["description"] == "A test domain"
    assert domain["domainInfo"]["name"] == "test-domain"
    assert domain["domainInfo"]["status"] == "REGISTERED"


@mock_aws
def test_describe_non_existent_domain_boto3():
    client = boto3.client("swf", region_name="us-west-1")

    with pytest.raises(ClientError) as ex:
        client.describe_domain(name="non-existent")
    assert ex.value.response["Error"]["Code"] == "UnknownResourceFault"
    assert ex.value.response["Error"]["Message"] == "Unknown domain: non-existent"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400



import boto3

from moto import mock_aws

test_ami = "/aws/service/ami-amazon-linux-latest/al2023-ami-kernel-default-x86_64"


@mock_aws
def test_ssm_get_latest_ami_by_path():
    client = boto3.client("ssm", region_name="us-west-1")
    path = "/aws/service/ami-amazon-linux-latest"
    params = client.get_parameters_by_path(Path=path)["Parameters"]
    assert len(params) == 10

    assert all(
        {p["Name"].startswith("/aws/service/ami-amazon-linux-latest") for p in params}
    )
    assert all({p["Type"] == "String" for p in params})
    assert all({p["DataType"] == "text" for p in params})
    assert all({p["ARN"].startswith("arn:aws:ssm:us-west-1") for p in params})


@mock_aws
def test_ssm_latest_amis_are_different_in_regions():
    client = boto3.client("ssm", region_name="us-west-1")
    ami_uswest = client.get_parameter(Name=test_ami)["Parameter"]["Value"]

    client = boto3.client("ssm", region_name="eu-north-1")
    ami_eunorth = client.get_parameter(Name=test_ami)["Parameter"]["Value"]

    assert ami_uswest != ami_eunorth



import abc

from moto.stepfunctions.parser.asl.component.intrinsic.argument.function_argument_list import (
    FunctionArgumentList,
)
from moto.stepfunctions.parser.asl.component.intrinsic.function.function import Function
from moto.stepfunctions.parser.asl.component.intrinsic.functionname.states_function_name import (
    StatesFunctionName,
)


class StatesFunction(Function, abc.ABC):
    name: StatesFunctionName

    def __init__(self, states_name: StatesFunctionName, arg_list: FunctionArgumentList):
        super().__init__(name=states_name, arg_list=arg_list)



import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws

from . import dynamodb_aws_verified


@mock_aws
def test_invalid_transact_get_items():
    dynamodb = boto3.resource("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test1",
        KeySchema=[{"AttributeName": "id", "KeyType": "HASH"}],
        AttributeDefinitions=[{"AttributeName": "id", "AttributeType": "S"}],
        ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
    )
    table = dynamodb.Table("test1")
    table.put_item(Item={"id": "1", "val": "1"})
    table.put_item(Item={"id": "1", "val": "2"})

    client = boto3.client("dynamodb", region_name="us-east-1")

    with pytest.raises(ClientError) as ex:
        client.transact_get_items(
            TransactItems=[
                {"Get": {"Key": {"id": {"S": "1"}}, "TableName": "test1"}}
                for i in range(26)
            ]
        )

    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert (
        "Member must have length less than or equal to 25"
        in ex.value.response["Error"]["Message"]
    )

    with pytest.raises(ClientError) as ex:
        client.transact_get_items(
            TransactItems=[
                {"Get": {"Key": {"id": {"S": "1"}}, "TableName": "test1"}},
                {"Get": {"Key": {"id": {"S": "1"}}, "TableName": "non_exists_table"}},
            ]
        )

    assert ex.value.response["Error"]["Code"] == "ResourceNotFoundException"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert ex.value.response["Error"]["Message"] == "Requested resource not found"


@mock_aws
def test_transact_get_items_should_return_empty_map_for_non_existent_item():
    client = boto3.client("dynamodb", region_name="us-west-2")
    table_name = "test-table"
    key_schema = [{"AttributeName": "id", "KeyType": "HASH"}]
    attribute_definitions = [{"AttributeName": "id", "AttributeType": "S"}]
    client.create_table(
        TableName=table_name,
        KeySchema=key_schema,
        AttributeDefinitions=attribute_definitions,
        ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
    )
    item = {"id": {"S": "1"}}
    client.put_item(TableName=table_name, Item=item)
    items = client.transact_get_items(
        TransactItems=[
            {"Get": {"Key": {"id": {"S": "1"}}, "TableName": table_name}},
            {"Get": {"Key": {"id": {"S": "2"}}, "TableName": table_name}},
        ]
    ).get("Responses", [])
    assert len(items) == 2
    assert items[0] == {"Item": item}
    assert items[1] == {}


@mock_aws
def test_valid_transact_get_items():
    dynamodb = boto3.resource("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test1",
        KeySchema=[
            {"AttributeName": "id", "KeyType": "HASH"},
            {"AttributeName": "sort_key", "KeyType": "RANGE"},
        ],
        AttributeDefinitions=[
            {"AttributeName": "id", "AttributeType": "S"},
            {"AttributeName": "sort_key", "AttributeType": "S"},
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
    )
    table1 = dynamodb.Table("test1")
    table1.put_item(Item={"id": "1", "sort_key": "1"})
    table1.put_item(Item={"id": "1", "sort_key": "2"})

    dynamodb.create_table(
        TableName="test2",
        KeySchema=[
            {"AttributeName": "id", "KeyType": "HASH"},
            {"AttributeName": "sort_key", "KeyType": "RANGE"},
        ],
        AttributeDefinitions=[
            {"AttributeName": "id", "AttributeType": "S"},
            {"AttributeName": "sort_key", "AttributeType": "S"},
        ],
        ProvisionedThroughput={"ReadCapacityUnits": 5, "WriteCapacityUnits": 5},
    )
    table2 = dynamodb.Table("test2")
    table2.put_item(Item={"id": "1", "sort_key": "1"})

    client = boto3.client("dynamodb", region_name="us-east-1")
    res = client.transact_get_items(
        TransactItems=[
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "1"}},
                    "TableName": "test1",
                }
            },
            {
                "Get": {
                    "Key": {"id": {"S": "non_exists_key"}, "sort_key": {"S": "2"}},
                    "TableName": "test1",
                }
            },
        ]
    )
    assert res["Responses"][0]["Item"] == {"id": {"S": "1"}, "sort_key": {"S": "1"}}
    assert len(res["Responses"]) == 2
    assert res["Responses"][1] == {}

    res = client.transact_get_items(
        TransactItems=[
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "1"}},
                    "TableName": "test1",
                }
            },
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "2"}},
                    "TableName": "test1",
                }
            },
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "1"}},
                    "TableName": "test2",
                }
            },
        ]
    )

    assert res["Responses"][0]["Item"] == {"id": {"S": "1"}, "sort_key": {"S": "1"}}

    assert res["Responses"][1]["Item"] == {"id": {"S": "1"}, "sort_key": {"S": "2"}}

    assert res["Responses"][2]["Item"] == {"id": {"S": "1"}, "sort_key": {"S": "1"}}

    res = client.transact_get_items(
        TransactItems=[
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "1"}},
                    "TableName": "test1",
                }
            },
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "2"}},
                    "TableName": "test1",
                }
            },
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "1"}},
                    "TableName": "test2",
                }
            },
        ],
        ReturnConsumedCapacity="TOTAL",
    )

    assert res["ConsumedCapacity"][0] == {
        "TableName": "test1",
        "CapacityUnits": 4.0,
        "ReadCapacityUnits": 4.0,
    }

    assert res["ConsumedCapacity"][1] == {
        "TableName": "test2",
        "CapacityUnits": 2.0,
        "ReadCapacityUnits": 2.0,
    }

    res = client.transact_get_items(
        TransactItems=[
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "1"}},
                    "TableName": "test1",
                }
            },
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "2"}},
                    "TableName": "test1",
                }
            },
            {
                "Get": {
                    "Key": {"id": {"S": "1"}, "sort_key": {"S": "1"}},
                    "TableName": "test2",
                }
            },
        ],
        ReturnConsumedCapacity="INDEXES",
    )

    assert res["ConsumedCapacity"][0] == {
        "TableName": "test1",
        "CapacityUnits": 4.0,
        "ReadCapacityUnits": 4.0,
        "Table": {"CapacityUnits": 4.0, "ReadCapacityUnits": 4.0},
    }

    assert res["ConsumedCapacity"][1] == {
        "TableName": "test2",
        "CapacityUnits": 2.0,
        "ReadCapacityUnits": 2.0,
        "Table": {"CapacityUnits": 2.0, "ReadCapacityUnits": 2.0},
    }


@mock_aws
def test_transact_write_items_put():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Put multiple items
    dynamodb.transact_write_items(
        TransactItems=[
            {
                "Put": {
                    "Item": {"id": {"S": f"foo{str(i)}"}, "foo": {"S": "bar"}},
                    "TableName": "test-table",
                }
            }
            for i in range(0, 5)
        ]
    )
    # Assert all are present
    items = dynamodb.scan(TableName="test-table")["Items"]
    assert len(items) == 5


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_transact_write_items_put_conditional_expressions(table_name=None):
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")

    dynamodb.put_item(TableName=table_name, Item={"pk": {"S": "foo2"}})
    # Put multiple items
    with pytest.raises(ClientError) as ex:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Put": {
                        "Item": {
                            "pk": {"S": f"foo{i}"},
                            "foo": {"S": "bar"},
                        },
                        "TableName": table_name,
                        "ConditionExpression": "#i <> :i",
                        "ExpressionAttributeNames": {"#i": "pk"},
                        "ExpressionAttributeValues": {
                            ":i": {
                                "S": "foo2"
                            }  # This item already exist, so the ConditionExpression should fail
                        },
                    }
                }
                for i in range(0, 5)
            ]
        )
    # Assert the exception is correct
    assert ex.value.response["Error"]["Code"] == "TransactionCanceledException"
    reasons = ex.value.response["CancellationReasons"]
    assert len(reasons) == 5
    assert {
        "Code": "ConditionalCheckFailed",
        "Message": "The conditional request failed",
    } in reasons
    assert {"Code": "None"} in reasons
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    # Assert all are present
    items = dynamodb.scan(TableName=table_name)["Items"]
    assert len(items) == 1
    assert items[0] == {"pk": {"S": "foo2"}}


@pytest.mark.aws_verified
@dynamodb_aws_verified()
def test_transact_write_items_failure__return_item(table_name=None):
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.put_item(TableName=table_name, Item={"pk": {"S": "foo2"}})
    # Put multiple items
    with pytest.raises(ClientError) as ex:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Put": {
                        "Item": {
                            "pk": {"S": f"foo{i}"},
                            "foo": {"S": "bar"},
                        },
                        "TableName": table_name,
                        "ConditionExpression": "#i <> :i",
                        "ExpressionAttributeNames": {"#i": "pk"},
                        # This man right here - should return item as part of error message
                        "ReturnValuesOnConditionCheckFailure": "ALL_OLD",
                        "ExpressionAttributeValues": {
                            ":i": {
                                "S": "foo2"
                            }  # This item already exist, so the ConditionExpression should fail
                        },
                    }
                }
                for i in range(0, 5)
            ]
        )
    # Assert the exception is correct
    assert ex.value.response["Error"]["Code"] == "TransactionCanceledException"
    reasons = ex.value.response["CancellationReasons"]
    assert len(reasons) == 5
    assert {
        "Code": "ConditionalCheckFailed",
        "Message": "The conditional request failed",
        "Item": {"pk": {"S": "foo2"}},
    } in reasons
    assert {"Code": "None"} in reasons
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    # Assert all are present
    items = dynamodb.scan(TableName=table_name)["Items"]
    assert len(items) == 1
    assert items[0] == {"pk": {"S": "foo2"}}


@mock_aws
def test_transact_write_items_conditioncheck_passes():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Insert an item with email address
    dynamodb.put_item(
        TableName="test-table",
        Item={"id": {"S": "foo"}, "email_address": {"S": "foo@moto.com"}},
    )
    # Put a new item, after verifying the exising item also has email address
    dynamodb.transact_write_items(
        TransactItems=[
            {
                "ConditionCheck": {
                    "Key": {"id": {"S": "foo"}},
                    "TableName": "test-table",
                    "ConditionExpression": "attribute_exists(#e)",
                    "ExpressionAttributeNames": {"#e": "email_address"},
                }
            },
            {
                "Put": {
                    "Item": {
                        "id": {"S": "bar"},
                        "email_address": {"S": "bar@moto.com"},
                    },
                    "TableName": "test-table",
                }
            },
        ]
    )
    # Assert all are present
    items = dynamodb.scan(TableName="test-table")["Items"]
    assert len(items) == 2
    assert {"email_address": {"S": "foo@moto.com"}, "id": {"S": "foo"}} in items
    assert {"email_address": {"S": "bar@moto.com"}, "id": {"S": "bar"}} in items


@mock_aws
def test_transact_write_items_conditioncheck_fails():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Insert an item without email address
    dynamodb.put_item(
        TableName="test-table",
        Item={"id": {"S": "foo"}},
    )
    # Try putting a new item, after verifying the exising item also has email address
    # ConditionCheck should fail
    with pytest.raises(ClientError) as ex:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "ConditionCheck": {
                        "Key": {"id": {"S": "foo"}},
                        "TableName": "test-table",
                        "ConditionExpression": "attribute_exists(#e)",
                        "ExpressionAttributeNames": {"#e": "email_address"},
                    }
                },
                {
                    "Put": {
                        "Item": {
                            "id": {"S": "bar"},
                            "email_address": {"S": "bar@moto.com"},
                        },
                        "TableName": "test-table",
                    }
                },
            ]
        )
    # Assert the exception is correct
    assert ex.value.response["Error"]["Code"] == "TransactionCanceledException"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400

    # Assert the original item is still present
    items = dynamodb.scan(TableName="test-table")["Items"]
    assert len(items) == 1
    assert items[0] == {"id": {"S": "foo"}}


@mock_aws
def test_transact_write_items_delete():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Insert an item
    dynamodb.put_item(TableName="test-table", Item={"id": {"S": "foo"}})
    # Delete the item
    dynamodb.transact_write_items(
        TransactItems=[
            {"Delete": {"Key": {"id": {"S": "foo"}}, "TableName": "test-table"}}
        ]
    )
    # Assert the item is deleted
    items = dynamodb.scan(TableName="test-table")["Items"]
    assert len(items) == 0


@mock_aws
def test_transact_write_items_delete_with_successful_condition_expression():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Insert an item without email address
    dynamodb.put_item(TableName="test-table", Item={"id": {"S": "foo"}})
    # ConditionExpression will pass - no email address has been specified yet
    dynamodb.transact_write_items(
        TransactItems=[
            {
                "Delete": {
                    "Key": {"id": {"S": "foo"}},
                    "TableName": "test-table",
                    "ConditionExpression": "attribute_not_exists(#e)",
                    "ExpressionAttributeNames": {"#e": "email_address"},
                }
            }
        ]
    )
    # Assert the item is deleted
    items = dynamodb.scan(TableName="test-table")["Items"]
    assert len(items) == 0


@mock_aws
def test_transact_write_items_delete_with_failed_condition_expression():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Insert an item with email address
    dynamodb.put_item(
        TableName="test-table",
        Item={"id": {"S": "foo"}, "email_address": {"S": "test@moto.com"}},
    )
    # Try to delete an item that does not have an email address
    # ConditionCheck should fail
    with pytest.raises(ClientError) as ex:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Delete": {
                        "Key": {"id": {"S": "foo"}},
                        "TableName": "test-table",
                        "ConditionExpression": "attribute_not_exists(#e)",
                        "ExpressionAttributeNames": {"#e": "email_address"},
                    }
                }
            ]
        )
    # Assert the exception is correct
    assert ex.value.response["Error"]["Code"] == "TransactionCanceledException"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    # Assert the original item is still present
    items = dynamodb.scan(TableName="test-table")["Items"]
    assert len(items) == 1
    assert items[0] == {"email_address": {"S": "test@moto.com"}, "id": {"S": "foo"}}


@mock_aws
def test_transact_write_items_update():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Insert an item
    dynamodb.put_item(TableName="test-table", Item={"id": {"S": "foo"}})
    # Update the item
    dynamodb.transact_write_items(
        TransactItems=[
            {
                "Update": {
                    "Key": {"id": {"S": "foo"}},
                    "TableName": "test-table",
                    "UpdateExpression": "SET #e = :v",
                    "ExpressionAttributeNames": {"#e": "email_address"},
                    "ExpressionAttributeValues": {":v": {"S": "test@moto.com"}},
                }
            }
        ]
    )
    # Assert the item is updated
    items = dynamodb.scan(TableName="test-table")["Items"]
    assert len(items) == 1
    assert items[0] == {"id": {"S": "foo"}, "email_address": {"S": "test@moto.com"}}


@mock_aws
def test_transact_write_items_update_with_failed_condition_expression():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Insert an item with email address
    dynamodb.put_item(
        TableName="test-table",
        Item={"id": {"S": "foo"}, "email_address": {"S": "test@moto.com"}},
    )
    # Try to update an item that does not have an email address
    # ConditionCheck should fail
    with pytest.raises(ClientError) as ex:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Update": {
                        "Key": {"id": {"S": "foo"}},
                        "TableName": "test-table",
                        "UpdateExpression": "SET #e = :v",
                        "ConditionExpression": "attribute_not_exists(#e)",
                        "ExpressionAttributeNames": {"#e": "email_address"},
                        "ExpressionAttributeValues": {":v": {"S": "update@moto.com"}},
                    }
                }
            ]
        )
    # Assert the exception is correct
    assert ex.value.response["Error"]["Code"] == "TransactionCanceledException"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    # Assert the original item is still present
    items = dynamodb.scan(TableName="test-table")["Items"]
    assert len(items) == 1
    assert items[0] == {"email_address": {"S": "test@moto.com"}, "id": {"S": "foo"}}


@mock_aws
def test_transact_write_items_fails_with_transaction_canceled_exception():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )
    # Insert one item
    dynamodb.put_item(TableName="test-table", Item={"id": {"S": "foo"}})
    # Update two items, the one that exists and another that doesn't
    with pytest.raises(ClientError) as ex:
        dynamodb.transact_write_items(
            TransactItems=[
                {
                    "Update": {
                        "Key": {"id": {"S": "foo"}},
                        "TableName": "test-table",
                        "UpdateExpression": "SET #k = :v",
                        "ConditionExpression": "attribute_exists(id)",
                        "ExpressionAttributeNames": {"#k": "key"},
                        "ExpressionAttributeValues": {":v": {"S": "value"}},
                    }
                },
                {
                    "Update": {
                        "Key": {"id": {"S": "doesnotexist"}},
                        "TableName": "test-table",
                        "UpdateExpression": "SET #e = :v",
                        "ConditionExpression": "attribute_exists(id)",
                        "ExpressionAttributeNames": {"#e": "key"},
                        "ExpressionAttributeValues": {":v": {"S": "value"}},
                    }
                },
            ]
        )
    assert ex.value.response["Error"]["Code"] == "TransactionCanceledException"
    assert ex.value.response["ResponseMetadata"]["HTTPStatusCode"] == 400
    assert (
        ex.value.response["Error"]["Message"]
        == "Transaction cancelled, please refer cancellation reasons for specific reasons [None, ConditionalCheckFailed]"
    )


@mock_aws
def test_transact_using_arns():
    table_schema = {
        "KeySchema": [{"AttributeName": "id", "KeyType": "HASH"}],
        "AttributeDefinitions": [{"AttributeName": "id", "AttributeType": "S"}],
    }
    dynamodb = boto3.client("dynamodb", region_name="us-east-1")
    table_arn = dynamodb.create_table(
        TableName="test-table", BillingMode="PAY_PER_REQUEST", **table_schema
    )["TableDescription"]["TableArn"]

    dynamodb.transact_write_items(
        TransactItems=[{"Put": {"Item": {"id": {"S": "foo"}}, "TableName": table_arn}}]
    )

    items = dynamodb.transact_get_items(
        TransactItems=[{"Get": {"Key": {"id": {"S": "foo"}}, "TableName": table_arn}}]
    )["Responses"]
    assert items == [{"Item": {"id": {"S": "foo"}}}]



"""Unit tests for sagemaker-supported APIs."""

import boto3

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


@mock_aws
def test_create_model_bias_job():
    client = boto3.client("sagemaker", region_name="us-west-2")

    response = client.create_model_bias_job_definition(
        JobDefinitionName="test-bias-job",
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/SageMakerRole",
        ModelBiasAppSpecification={
            "ImageUri": f"{ACCOUNT_ID}.dkr.ecr.us-west-2.amazonaws.com/my-image:latest",
            "ConfigUri": "s3://my-bucket/bias-config.json",
        },
        ModelBiasJobInput={
            "EndpointInput": {
                "EndpointName": "test-endpoint",
                "LocalPath": "/opt/ml/processing/input/endpoint",
                "S3InputMode": "File",
                "S3DataDistributionType": "FullyReplicated",
            },
            "GroundTruthS3Input": {
                "S3Uri": "s3://my-bucket/ground-truth-data",
            },
        },
        ModelBiasJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://my-bucket/bias-output",
                        "LocalPath": "/opt/ml/processing/output",
                        "S3UploadMode": "EndOfJob",
                    },
                },
            ],
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m5.xlarge",
                "VolumeSizeInGB": 10,
            },
        },
        NetworkConfig={
            "EnableInterContainerTrafficEncryption": False,
            "EnableNetworkIsolation": False,
        },
    )

    expected_job_definition_arn = f"arn:aws:sagemaker:us-west-2:{ACCOUNT_ID}:model-bias-job-definition/test-bias-job"
    assert response["JobDefinitionArn"] == expected_job_definition_arn


@mock_aws
def test_describe_model_bias_job():
    client = boto3.client("sagemaker", region_name="us-east-1")

    job_name = "test-model-bias-job"
    client.create_model_bias_job_definition(
        JobDefinitionName=job_name,
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/SageMakerRole",
        ModelBiasAppSpecification={
            "ImageUri": f"{ACCOUNT_ID}.dkr.ecr.us-west-2.amazonaws.com/my-image:latest",
            "ConfigUri": "s3://my-bucket/bias-config.json",
        },
        ModelBiasJobInput={
            "EndpointInput": {
                "EndpointName": "test-endpoint",
                "LocalPath": "/opt/ml/processing/input/endpoint",
                "S3InputMode": "File",
                "S3DataDistributionType": "FullyReplicated",
            },
            "GroundTruthS3Input": {
                "S3Uri": "s3://my-bucket/ground-truth-data",
            },
        },
        ModelBiasJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://my-bucket/bias-output",
                        "LocalPath": "/opt/ml/processing/output",
                        "S3UploadMode": "EndOfJob",
                    },
                },
            ],
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m5.xlarge",
                "VolumeSizeInGB": 10,
            },
        },
        NetworkConfig={
            "EnableInterContainerTrafficEncryption": False,
            "EnableNetworkIsolation": False,
        },
    )
    response = client.describe_model_bias_job_definition(JobDefinitionName=job_name)
    assert response["JobDefinitionName"] == job_name
    assert (
        response["JobDefinitionArn"]
        == f"arn:aws:sagemaker:us-east-1:{ACCOUNT_ID}:model-bias-job-definition/{job_name}"
    )


@mock_aws
def test_list_model_bias_jobs():
    client = boto3.client("sagemaker", region_name="us-east-1")
    client.create_model_bias_job_definition(
        JobDefinitionName="test-bias-job-1",
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/SageMakerRole",
        ModelBiasAppSpecification={
            "ImageUri": f"{ACCOUNT_ID}.dkr.ecr.us-west-2.amazonaws.com/my-image:latest",
            "ConfigUri": "s3://my-bucket/bias-config.json",
        },
        ModelBiasJobInput={
            "EndpointInput": {
                "EndpointName": "test-endpoint",
                "LocalPath": "/opt/ml/processing/input/endpoint",
                "S3InputMode": "File",
                "S3DataDistributionType": "FullyReplicated",
            },
            "GroundTruthS3Input": {
                "S3Uri": "s3://my-bucket/ground-truth-data",
            },
        },
        ModelBiasJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://my-bucket/bias-output",
                        "LocalPath": "/opt/ml/processing/output",
                        "S3UploadMode": "EndOfJob",
                    },
                },
            ],
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m5.xlarge",
                "VolumeSizeInGB": 10,
            },
        },
        NetworkConfig={
            "EnableInterContainerTrafficEncryption": False,
            "EnableNetworkIsolation": False,
        },
    )
    client.create_model_bias_job_definition(
        JobDefinitionName="test-bias-job-2",
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/SageMakerRole",
        ModelBiasAppSpecification={
            "ImageUri": f"{ACCOUNT_ID}.dkr.ecr.us-west-2.amazonaws.com/my-image:latest",
            "ConfigUri": "s3://my-bucket/bias-config.json",
        },
        ModelBiasJobInput={
            "EndpointInput": {
                "EndpointName": "test-endpoint",
                "LocalPath": "/opt/ml/processing/input/endpoint",
                "S3InputMode": "File",
                "S3DataDistributionType": "FullyReplicated",
            },
            "GroundTruthS3Input": {
                "S3Uri": "s3://my-bucket/ground-truth-data",
            },
        },
        ModelBiasJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://my-bucket/bias-output",
                        "LocalPath": "/opt/ml/processing/output",
                        "S3UploadMode": "EndOfJob",
                    },
                },
            ],
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m5.xlarge",
                "VolumeSizeInGB": 10,
            },
        },
        NetworkConfig={
            "EnableInterContainerTrafficEncryption": False,
            "EnableNetworkIsolation": False,
        },
    )
    response = client.list_model_bias_job_definitions()

    job_names = [
        job["MonitoringJobDefinitionName"] for job in response["JobDefinitionSummaries"]
    ]
    assert "test-bias-job-1" in job_names
    assert "test-bias-job-2" in job_names


@mock_aws
def test_delete_model_bias_job():
    client = boto3.client("sagemaker", region_name="us-east-1")

    client.create_model_bias_job_definition(
        JobDefinitionName="test-bias-job",
        RoleArn=f"arn:aws:iam::{ACCOUNT_ID}:role/SageMakerRole",
        ModelBiasAppSpecification={
            "ImageUri": f"{ACCOUNT_ID}.dkr.ecr.us-west-2.amazonaws.com/my-image:latest",
            "ConfigUri": "s3://my-bucket/bias-config.json",
        },
        ModelBiasJobInput={
            "EndpointInput": {
                "EndpointName": "test-endpoint",
                "LocalPath": "/opt/ml/processing/input/endpoint",
                "S3InputMode": "File",
                "S3DataDistributionType": "FullyReplicated",
            },
            "GroundTruthS3Input": {
                "S3Uri": "s3://my-bucket/ground-truth-data",
            },
        },
        ModelBiasJobOutputConfig={
            "MonitoringOutputs": [
                {
                    "S3Output": {
                        "S3Uri": "s3://my-bucket/bias-output",
                        "LocalPath": "/opt/ml/processing/output",
                        "S3UploadMode": "EndOfJob",
                    },
                },
            ],
        },
        JobResources={
            "ClusterConfig": {
                "InstanceCount": 1,
                "InstanceType": "ml.m5.xlarge",
                "VolumeSizeInGB": 10,
            },
        },
        NetworkConfig={
            "EnableInterContainerTrafficEncryption": False,
            "EnableNetworkIsolation": False,
        },
    )
    client.delete_model_bias_job_definition(JobDefinitionName="test-bias-job")
    response = client.list_model_bias_job_definitions()
    job_names = [job["JobDefinitionName"] for job in response["JobDefinitionSummaries"]]
    assert "test-bias-job" not in job_names



import boto3

from moto import mock_aws
from moto.core.decorator import MockAWS


@mock_aws
def method_without_parentheses() -> int:
    assert boto3.client("s3").list_buckets()["Buckets"] == []
    return 123


@mock_aws()
def method_with_parentheses() -> int:
    assert boto3.client("s3").list_buckets()["Buckets"] == []
    return 456


@mock_aws
def test_no_return() -> None:
    assert boto3.client("s3").list_buckets()["Buckets"] == []


def test_with_context_manager() -> None:
    with mock_aws():
        assert boto3.client("s3").list_buckets()["Buckets"] == []


def test_manual() -> None:
    # this has the explicit type not because it's necessary but so that mypy will
    # complain if it's wrong
    m: MockAWS = mock_aws()
    m.start()
    assert boto3.client("s3").list_buckets()["Buckets"] == []
    m.stop()


x: int = method_with_parentheses()
assert x == 456

y: int = method_without_parentheses()
assert y == 123



import boto3

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID
from moto.moto_api._internal import mock_random

from .test_firehose_destination_types import create_http_delivery_stream

TEST_REGION = "us-west-1"


@mock_aws
def test_create_http_stream():
    """Verify fields of a HTTP delivery stream."""
    client = boto3.client("firehose", region_name=TEST_REGION)

    stream_name = f"stream_{mock_random.get_random_hex(6)}"
    response = create_http_delivery_stream(client, stream_name)
    stream_arn = response["DeliveryStreamARN"]

    response = client.describe_delivery_stream(DeliveryStreamName=stream_name)
    stream_description = response["DeliveryStreamDescription"]

    # Sure and Freezegun don't play nicely together
    stream_description.pop("CreateTimestamp")
    stream_description.pop("LastUpdateTimestamp")

    assert stream_description == {
        "DeliveryStreamName": stream_name,
        "DeliveryStreamARN": stream_arn,
        "DeliveryStreamStatus": "ACTIVE",
        "DeliveryStreamType": "DirectPut",
        "VersionId": "1",
        "Destinations": [
            {
                "DestinationId": "destinationId-000000000001",
                "HttpEndpointDestinationDescription": {
                    "EndpointConfiguration": {"Url": "google.com"},
                    "RetryOptions": {"DurationInSeconds": 100},
                    "BufferingHints": {"SizeInMBs": 123, "IntervalInSeconds": 124},
                    "CloudWatchLoggingOptions": {"Enabled": False},
                    "S3DestinationDescription": {
                        "RoleARN": f"arn:aws:iam::{ACCOUNT_ID}:role/firehose_delivery_role",
                        "BucketARN": "arn:aws:s3:::firehose-test",
                        "Prefix": "myFolder/",
                        "BufferingHints": {
                            "SizeInMBs": 123,
                            "IntervalInSeconds": 124,
                        },
                        "CompressionFormat": "UNCOMPRESSED",
                    },
                },
            }
        ],
        "HasMoreDestinations": False,
    }


@mock_aws
def test_update_s3_for_http_stream():
    """Verify fields of a HTTP delivery stream."""
    client = boto3.client("firehose", region_name=TEST_REGION)

    stream_name = f"stream_{mock_random.get_random_hex(6)}"
    create_http_delivery_stream(client, stream_name)

    stream = client.describe_delivery_stream(DeliveryStreamName=stream_name)
    version_id = stream["DeliveryStreamDescription"]["VersionId"]

    client.update_destination(
        DeliveryStreamName=stream_name,
        CurrentDeliveryStreamVersionId=version_id,
        DestinationId="destinationId-000000000001",
        HttpEndpointDestinationUpdate={"S3Update": {"ErrorOutputPrefix": "prefix2"}},
    )

    desc = client.describe_delivery_stream(DeliveryStreamName=stream_name)[
        "DeliveryStreamDescription"
    ]
    s3_desc = desc["Destinations"][0]["HttpEndpointDestinationDescription"][
        "S3DestinationDescription"
    ]
    assert s3_desc["ErrorOutputPrefix"] == "prefix2"



import weakref
from collections import defaultdict
from typing import Any, Dict, Iterator, List, Optional

from moto.core.common_models import CloudFormationModel

from ..exceptions import (
    InvalidVPCPeeringConnectionIdError,
    InvalidVPCPeeringConnectionStateTransitionError,
    OperationNotPermitted2,
    OperationNotPermitted3,
    OperationNotPermitted5,
)
from ..utils import random_vpc_peering_connection_id
from .core import TaggedEC2Resource
from .vpcs import VPC


class PeeringConnectionStatus:
    def __init__(
        self, accepter_id: str, code: str = "initiating-request", message: str = ""
    ):
        self.accepter_id = accepter_id
        self.code = code
        self.message = message

    def deleted(self, deleter_id: str) -> None:
        self.code = "deleted"
        self.message = f"Deleted by {deleter_id}"

    def initiating(self) -> None:
        self.code = "initiating-request"
        self.message = f"Initiating Request to {self.accepter_id}"

    def pending(self) -> None:
        self.code = "pending-acceptance"
        self.message = f"Pending Acceptance by {self.accepter_id}"

    def accept(self) -> None:
        self.code = "active"
        self.message = "Active"

    def reject(self) -> None:
        self.code = "rejected"
        self.message = "Inactive"


class VPCPeeringConnection(TaggedEC2Resource, CloudFormationModel):
    DEFAULT_OPTIONS = {
        "AllowEgressFromLocalClassicLinkToRemoteVpc": "false",
        "AllowEgressFromLocalVpcToRemoteClassicLink": "false",
        "AllowDnsResolutionFromRemoteVpc": "false",
    }

    def __init__(
        self,
        backend: Any,
        vpc_pcx_id: str,
        vpc: VPC,
        peer_vpc: VPC,
        tags: Optional[Dict[str, str]] = None,
    ):
        self.id = vpc_pcx_id
        self.ec2_backend = backend
        self.vpc = vpc
        self.peer_vpc = peer_vpc
        self.requester_options = self.DEFAULT_OPTIONS.copy()
        self.accepter_options = self.DEFAULT_OPTIONS.copy()
        self.add_tags(tags or {})
        self._status = PeeringConnectionStatus(accepter_id=peer_vpc.owner_id)

    @staticmethod
    def cloudformation_name_type() -> str:
        return ""

    @staticmethod
    def cloudformation_type() -> str:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-vpcpeeringconnection.html
        return "AWS::EC2::VPCPeeringConnection"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "VPCPeeringConnection":
        from ..models import ec2_backends

        properties = cloudformation_json["Properties"]

        ec2_backend = ec2_backends[account_id][region_name]
        vpc = ec2_backend.get_vpc(properties["VpcId"])
        peer_vpc = ec2_backend.get_vpc(properties["PeerVpcId"])

        vpc_pcx = ec2_backend.create_vpc_peering_connection(vpc, peer_vpc)

        return vpc_pcx

    @property
    def physical_resource_id(self) -> str:
        return self.id


class VPCPeeringConnectionBackend:
    # for cross region vpc reference
    vpc_pcx_refs = defaultdict(set)  # type: ignore

    def __init__(self) -> None:
        self.vpc_pcxs: Dict[str, VPCPeeringConnection] = {}
        self.vpc_pcx_refs[self.__class__].add(weakref.ref(self))

    @classmethod
    def get_vpc_pcx_refs(cls) -> Iterator[VPCPeeringConnection]:
        for inst_ref in cls.vpc_pcx_refs[cls]:
            inst = inst_ref()
            if inst is not None:
                yield inst

    def create_vpc_peering_connection(
        self, vpc: VPC, peer_vpc: VPC, tags: Optional[Dict[str, str]] = None
    ) -> VPCPeeringConnection:
        vpc_pcx_id = random_vpc_peering_connection_id()
        vpc_pcx = VPCPeeringConnection(self, vpc_pcx_id, vpc, peer_vpc, tags)
        vpc_pcx._status.pending()
        self.vpc_pcxs[vpc_pcx_id] = vpc_pcx
        # insert cross-account/cross-region peering info
        if vpc.owner_id != peer_vpc.owner_id or vpc.region != peer_vpc.region:
            for backend in peer_vpc.ec2_backend.get_vpc_pcx_refs():
                if (
                    backend.account_id == peer_vpc.owner_id
                    and backend.region_name == peer_vpc.region
                ):
                    backend.vpc_pcxs[vpc_pcx_id] = vpc_pcx

        return vpc_pcx

    def describe_vpc_peering_connections(
        self, vpc_peering_ids: Optional[List[str]] = None
    ) -> List[VPCPeeringConnection]:
        all_pcxs = list(self.vpc_pcxs.values())
        if vpc_peering_ids:
            return [pcx for pcx in all_pcxs if pcx.id in vpc_peering_ids]
        return all_pcxs

    def get_vpc_peering_connection(self, vpc_pcx_id: str) -> VPCPeeringConnection:
        if vpc_pcx_id not in self.vpc_pcxs:
            raise InvalidVPCPeeringConnectionIdError(vpc_pcx_id)
        return self.vpc_pcxs[vpc_pcx_id]

    def delete_vpc_peering_connection(self, vpc_pcx_id: str) -> VPCPeeringConnection:
        deleted = self.get_vpc_peering_connection(vpc_pcx_id)
        deleted._status.deleted(deleter_id=self.account_id)  # type: ignore[attr-defined]
        return deleted

    def accept_vpc_peering_connection(self, vpc_pcx_id: str) -> VPCPeeringConnection:
        vpc_pcx = self.get_vpc_peering_connection(vpc_pcx_id)

        # validate cross-account acceptance
        req_account_id = vpc_pcx.vpc.owner_id
        acp_account_id = vpc_pcx.peer_vpc.owner_id
        if req_account_id != acp_account_id and self.account_id != acp_account_id:  # type: ignore[attr-defined]
            raise OperationNotPermitted5(self.account_id, vpc_pcx_id, "accept")  # type: ignore[attr-defined]

        # validate cross-region acceptance
        pcx_req_region = vpc_pcx.vpc.region
        pcx_acp_region = vpc_pcx.peer_vpc.region
        if pcx_req_region != pcx_acp_region and self.region_name == pcx_req_region:  # type: ignore[attr-defined]
            raise OperationNotPermitted2(self.region_name, vpc_pcx.id, pcx_acp_region)  # type: ignore[attr-defined]

        if vpc_pcx._status.code != "pending-acceptance":
            raise InvalidVPCPeeringConnectionStateTransitionError(vpc_pcx.id)
        vpc_pcx._status.accept()
        return vpc_pcx

    def reject_vpc_peering_connection(self, vpc_pcx_id: str) -> VPCPeeringConnection:
        vpc_pcx = self.get_vpc_peering_connection(vpc_pcx_id)

        # validate cross-account rejection
        req_account_id = vpc_pcx.vpc.owner_id
        acp_account_id = vpc_pcx.peer_vpc.owner_id
        if req_account_id != acp_account_id and self.account_id != acp_account_id:  # type: ignore[attr-defined]
            raise OperationNotPermitted5(self.account_id, vpc_pcx_id, "reject")  # type: ignore[attr-defined]

        # validate cross-region acceptance
        pcx_req_region = vpc_pcx.vpc.region
        pcx_acp_region = vpc_pcx.peer_vpc.region
        if pcx_req_region != pcx_acp_region and self.region_name == pcx_req_region:  # type: ignore[attr-defined]
            raise OperationNotPermitted3(self.region_name, vpc_pcx.id, pcx_acp_region)  # type: ignore[attr-defined]

        if vpc_pcx._status.code != "pending-acceptance":
            raise InvalidVPCPeeringConnectionStateTransitionError(vpc_pcx.id)
        vpc_pcx._status.reject()
        return vpc_pcx

    def modify_vpc_peering_connection_options(
        self,
        vpc_pcx_id: str,
        accepter_options: Optional[Dict[str, Any]] = None,
        requester_options: Optional[Dict[str, Any]] = None,
    ) -> None:
        vpc_pcx = self.get_vpc_peering_connection(vpc_pcx_id)
        if not vpc_pcx:
            raise InvalidVPCPeeringConnectionIdError(vpc_pcx_id)
        # TODO: check if actual vpc has this options enabled
        if accepter_options:
            vpc_pcx.accepter_options.update(accepter_options)
        if requester_options:
            vpc_pcx.requester_options.update(requester_options)



from itertools import chain
from typing import Final

from moto.stepfunctions.parser.asl.component.component import Component


class Label(Component):
    label: Final[str]

    def __init__(self, label: str):
        self.label = label.encode().decode("unicode-escape")

        if len(self.label) == 0:
            raise ValueError("Label cannot be empty")

        if len(self.label) > 40:
            raise ValueError("Label cannot exceed 40 characters")

        for invalid_char in list(' ?*<>{}[]:;,\\|^~$#%&`"') + [
            chr(i) for i in chain(range(0x00, 0x20), range(0x7F, 0xA0))
        ]:
            if invalid_char in self.label:
                escaped_char = invalid_char.encode("unicode-escape").decode()
                raise ValueError(f"Label contains invalid character: '{escaped_char}'")



"""EBSBackend class with methods for supported APIs."""

from typing import Any, Dict, List, Optional, Tuple

from moto.core.base_backend import BackendDict, BaseBackend
from moto.core.common_models import BaseModel
from moto.core.utils import unix_time
from moto.ec2.models import EC2Backend, ec2_backends
from moto.ec2.models.elastic_block_store import Snapshot
from moto.moto_api._internal import mock_random


class Block(BaseModel):
    def __init__(
        self, block_data: str, checksum: str, checksum_algorithm: str, data_length: str
    ):
        self.block_data = block_data
        self.checksum = checksum
        self.checksum_algorithm = checksum_algorithm
        self.data_length = data_length
        self.block_token = str(mock_random.uuid4())


class EBSSnapshot(BaseModel):
    def __init__(self, account_id: str, snapshot: Snapshot):
        self.account_id = account_id
        self.snapshot_id = snapshot.id
        self.status = "pending"
        self.start_time = unix_time()
        self.volume_size = snapshot.volume.size
        self.block_size = 512
        self.tags = [
            {"Key": t["key"], "Value": t["value"]} for t in snapshot.get_tags()
        ]
        self.description = snapshot.description

        self.blocks: Dict[str, Block] = dict()

    def put_block(
        self,
        block_idx: str,
        block_data: str,
        checksum: str,
        checksum_algorithm: str,
        data_length: str,
    ) -> None:
        block = Block(block_data, checksum, checksum_algorithm, data_length)
        self.blocks[block_idx] = block

    def to_json(self) -> Dict[str, Any]:
        return {
            "SnapshotId": self.snapshot_id,
            "OwnerId": self.account_id,
            "Status": self.status,
            "StartTime": self.start_time,
            "VolumeSize": self.volume_size,
            "BlockSize": self.block_size,
            "Tags": self.tags,
            "Description": self.description,
        }


class EBSBackend(BaseBackend):
    """Implementation of EBS APIs."""

    def __init__(self, region_name: str, account_id: str):
        super().__init__(region_name, account_id)
        self.snapshots: Dict[str, EBSSnapshot] = dict()

    @property
    def ec2_backend(self) -> EC2Backend:
        return ec2_backends[self.account_id][self.region_name]

    def start_snapshot(
        self, volume_size: int, tags: Optional[List[Dict[str, str]]], description: str
    ) -> EBSSnapshot:
        """
        The following parameters are not yet implemented: ParentSnapshotId, ClientToken, Encrypted, KmsKeyArn, Timeout
        """
        zone_name = f"{self.region_name}a"
        vol = self.ec2_backend.create_volume(size=volume_size, zone_name=zone_name)
        snapshot = self.ec2_backend.create_snapshot(
            volume_id=vol.id, description=description
        )
        if tags:
            snapshot.add_tags({tag["Key"]: tag["Value"] for tag in tags})
        ebs_snapshot = EBSSnapshot(account_id=self.account_id, snapshot=snapshot)
        self.snapshots[ebs_snapshot.snapshot_id] = ebs_snapshot
        return ebs_snapshot

    def complete_snapshot(self, snapshot_id: str) -> Dict[str, str]:
        """
        The following parameters are not yet supported: ChangedBlocksCount, Checksum, ChecksumAlgorithm, ChecksumAggregationMethod
        """
        self.snapshots[snapshot_id].status = "completed"
        return {"Status": "completed"}

    def put_snapshot_block(
        self,
        snapshot_id: str,
        block_index: str,
        block_data: str,
        checksum: str,
        checksum_algorithm: str,
        data_length: str,
    ) -> Tuple[str, str]:
        """
        The following parameters are currently not taken into account: DataLength, Progress.
        The Checksum and ChecksumAlgorithm are taken at face-value, but no validation takes place.
        """
        snapshot = self.snapshots[snapshot_id]
        snapshot.put_block(
            block_index, block_data, checksum, checksum_algorithm, data_length
        )
        return checksum, checksum_algorithm

    def get_snapshot_block(self, snapshot_id: str, block_index: str) -> Block:
        """
        The BlockToken-parameter is not yet implemented
        """
        snapshot = self.snapshots[snapshot_id]
        return snapshot.blocks[block_index]

    def list_changed_blocks(
        self, first_snapshot_id: str, second_snapshot_id: str
    ) -> Tuple[Dict[str, Tuple[str, Optional[str]]], EBSSnapshot]:
        """
        The following parameters are not yet implemented: NextToken, MaxResults, StartingBlockIndex
        """
        snapshot1 = self.snapshots[first_snapshot_id]
        snapshot2 = self.snapshots[second_snapshot_id]
        changed_blocks: Dict[
            str, Tuple[str, Optional[str]]
        ] = {}  # {idx: (token1, token2), ..}
        for idx in snapshot1.blocks:
            block1 = snapshot1.blocks[idx]
            if idx in snapshot2.blocks:
                block2 = snapshot2.blocks[idx]
                if block1.block_data != block2.block_data:
                    changed_blocks[idx] = (block1.block_token, block2.block_token)
            else:
                changed_blocks[idx] = (block1.block_token, None)

        return changed_blocks, snapshot1

    def list_snapshot_blocks(self, snapshot_id: str) -> EBSSnapshot:
        """
        The following parameters are not yet implemented: NextToken, MaxResults, StartingBlockIndex
        """
        return self.snapshots[snapshot_id]


ebs_backends = BackendDict(EBSBackend, "ebs")



from collections import OrderedDict
from typing import Any, Dict, List, Optional

from moto.core.common_models import CloudFormationModel

from ..exceptions import (
    InvalidLaunchTemplateNameAlreadyExistsError,
    InvalidLaunchTemplateNameNotFoundError,
    InvalidLaunchTemplateNameNotFoundWithNameError,
    MissingParameterError,
)
from ..utils import (
    convert_tag_spec,
    generic_filter,
    random_launch_template_id,
    random_launch_template_name,
    utc_date_and_time,
)
from .core import TaggedEC2Resource


class LaunchTemplateVersion:
    def __init__(
        self,
        template: "LaunchTemplate",
        number: int,
        data: Dict[str, Any],
        description: str,
    ):
        self.template = template
        self.number = number
        self.data = data
        self.description = description
        self.create_time = utc_date_and_time()
        self.instance_tags = convert_tag_spec(data.get("TagSpecification", [])).get(
            "instance", {}
        )

    @property
    def image_id(self) -> str:
        return self.data.get("ImageId", "")

    @property
    def instance_type(self) -> str:
        return self.data.get("InstanceType", "")

    @property
    def security_groups(self) -> List[str]:
        return self.data.get("SecurityGroups", [])

    @property
    def user_data(self) -> str:
        return self.data.get("UserData", "")


class LaunchTemplate(TaggedEC2Resource, CloudFormationModel):
    def __init__(
        self,
        backend: Any,
        name: str,
        template_data: Dict[str, Any],
        version_description: str,
        tag_spec: Dict[str, Dict[str, str]],
    ):
        self.ec2_backend = backend
        self.name = name
        self.id = random_launch_template_id()
        self.create_time = utc_date_and_time()
        tag_map: Dict[str, str] = tag_spec.get("launch-template", {})
        self.add_tags(tag_map)
        self.tags = self.get_tags()

        self.versions: List[LaunchTemplateVersion] = []
        self.create_version(template_data, version_description)
        self.default_version_number = 1

    def create_version(
        self, data: Dict[str, Any], description: str
    ) -> LaunchTemplateVersion:
        num = len(self.versions) + 1
        version = LaunchTemplateVersion(self, num, data, description)
        self.versions.append(version)
        return version

    def is_default(self, version: LaunchTemplateVersion) -> bool:
        return self.default_version == version.number  # type: ignore

    def get_version(self, num: Any) -> LaunchTemplateVersion:
        if str(num).lower() == "$latest":
            return self.versions[-1]
        if str(num).lower() == "$default":
            return self.default_version()
        return self.versions[int(num) - 1]

    def default_version(self) -> LaunchTemplateVersion:
        return self.versions[self.default_version_number - 1]

    def latest_version(self) -> LaunchTemplateVersion:
        return self.versions[-1]

    @property
    def latest_version_number(self) -> int:
        return self.latest_version().number

    @property
    def physical_resource_id(self) -> str:
        return self.id

    def get_filter_value(
        self, filter_name: str, method_name: Optional[str] = None
    ) -> Any:
        if filter_name == "launch-template-name":
            return self.name
        else:
            return super().get_filter_value(filter_name, "DescribeLaunchTemplates")

    @staticmethod
    def cloudformation_name_type() -> str:
        return "LaunchTemplateName"

    @staticmethod
    def cloudformation_type() -> str:
        # https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-resource-ec2-launchtemplate.html
        return "AWS::EC2::LaunchTemplate"

    @classmethod
    def create_from_cloudformation_json(  # type: ignore[misc]
        cls,
        resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
        **kwargs: Any,
    ) -> "LaunchTemplate":
        from ..models import ec2_backends

        backend = ec2_backends[account_id][region_name]

        properties = cloudformation_json["Properties"]
        name = properties.get("LaunchTemplateName")
        data = properties.get("LaunchTemplateData")
        description = properties.get("VersionDescription")
        tag_spec = convert_tag_spec(
            properties.get("TagSpecifications", {}), tag_key="Tags"
        )

        if name is None:
            name = random_launch_template_name()

        launch_template = backend.create_launch_template(
            name, description, data, tag_spec
        )

        return launch_template

    @classmethod
    def update_from_cloudformation_json(  # type: ignore[misc]
        cls,
        original_resource: Any,
        new_resource_name: str,
        cloudformation_json: Any,
        account_id: str,
        region_name: str,
    ) -> "LaunchTemplate":
        from ..models import ec2_backends

        backend = ec2_backends[account_id][region_name]

        properties = cloudformation_json["Properties"]

        data = properties.get("LaunchTemplateData")
        description = properties.get("VersionDescription")

        launch_template = backend.get_launch_template(original_resource.id)

        launch_template.create_version(data, description)

        return launch_template

    def delete(
        self,
        account_id: str,
        region_name: str,
    ) -> None:
        from ..models import ec2_backends

        backend = ec2_backends[account_id][region_name]
        backend.delete_launch_template(self.name, None)

    @classmethod
    def has_cfn_attr(cls, attr: str) -> bool:
        return attr in [
            "DefaultVersionNumber",
            "LaunchTemplateId",
            "LatestVersionNumber",
        ]

    def get_cfn_attribute(self, attribute_name: str) -> str:
        from moto.cloudformation.exceptions import UnformattedGetAttTemplateException

        if attribute_name == "DefaultVersionNumber":
            return str(self.default_version_number)
        if attribute_name == "LaunchTemplateId":
            return self.id
        if attribute_name == "LatestVersionNumber":
            return str(self.latest_version_number)
        raise UnformattedGetAttTemplateException()


class LaunchTemplateBackend:
    def __init__(self) -> None:
        self.launch_template_name_to_ids: Dict[str, str] = {}
        self.launch_templates: Dict[str, LaunchTemplate] = OrderedDict()
        self.launch_template_insert_order: List[str] = []

    def create_launch_template(
        self,
        name: str,
        description: str,
        template_data: Dict[str, Any],
        tag_spec: Dict[str, Any],
    ) -> LaunchTemplate:
        if name in self.launch_template_name_to_ids:
            raise InvalidLaunchTemplateNameAlreadyExistsError()
        template = LaunchTemplate(self, name, template_data, description, tag_spec)
        self.launch_templates[template.id] = template
        self.launch_template_name_to_ids[template.name] = template.id
        self.launch_template_insert_order.append(template.id)
        return template

    def get_launch_template(self, template_id: str) -> LaunchTemplate:
        return self.launch_templates[template_id]

    def get_launch_template_by_name(self, name: str) -> LaunchTemplate:
        if name not in self.launch_template_name_to_ids:
            raise InvalidLaunchTemplateNameNotFoundWithNameError(name)
        return self.get_launch_template(self.launch_template_name_to_ids[name])

    def delete_launch_template(self, name: str, tid: Optional[str]) -> LaunchTemplate:
        if name:
            tid = self.launch_template_name_to_ids.get(name)
        if tid is None:
            raise MissingParameterError("launch template ID or launch template name")
        if tid not in self.launch_templates:
            raise InvalidLaunchTemplateNameNotFoundError()
        template = self.launch_templates.pop(tid)
        self.launch_template_name_to_ids.pop(template.name, None)
        return template

    def describe_launch_templates(
        self,
        template_names: Optional[List[str]] = None,
        template_ids: Optional[List[str]] = None,
        filters: Any = None,
    ) -> List[LaunchTemplate]:
        if template_names and not template_ids:
            template_ids = []
            for name in template_names:
                if name not in self.launch_template_name_to_ids:
                    raise InvalidLaunchTemplateNameNotFoundError()
                template_ids.append(self.launch_template_name_to_ids[name])

        if template_ids:
            templates = [
                self.launch_templates[tid]
                for tid in template_ids
                if tid in self.launch_templates
            ]
        else:
            templates = list(self.launch_templates.values())

        return generic_filter(filters, templates)

    def get_launch_template_data(self, instance_id: str) -> Any:
        return self.get_instance(instance_id)  # type: ignore[attr-defined]



"""sagemakerruntime base URL and path."""

from .responses import SageMakerRuntimeResponse

url_bases = [
    r"https?://runtime\.sagemaker\.(.+)\.amazonaws\.com",
]


response = SageMakerRuntimeResponse()


url_paths = {
    "{0}/endpoints/(?P<name>[^/]+)/async-invocations$": response.dispatch,
    "{0}/endpoints/(?P<name>[^/]+)/invocations$": response.dispatch,
}



"""appmesh base URL and path."""

from .responses import AppMeshResponse

url_bases = [
    r"https?://appmesh\.(.+)\.amazonaws\.com",
]

url_paths = {
    "{0}/v20190125/meshes$": AppMeshResponse.dispatch,
    "{0}/v20190125/meshes/(?P<meshName>[^/]+)$": AppMeshResponse.dispatch,
    "{0}/v20190125/tags$": AppMeshResponse.dispatch,
    "{0}/v20190125/tag$": AppMeshResponse.dispatch,
    "{0}/v20190125/meshes/(?P<meshName>.*)/virtualRouters/(?P<virtualRouterName>[^/]+)$": AppMeshResponse.dispatch,
    "{0}/v20190125/meshes/(?P<meshName>.*)/virtualRouters$": AppMeshResponse.dispatch,
    "{0}/v20190125/meshes/(?P<meshName>.*)/virtualRouter/(?P<virtualRouterName>.*)/routes$": AppMeshResponse.dispatch,
    "{0}/v20190125/meshes/(?P<meshName>.*)/virtualRouter/(?P<virtualRouterName>.*)/routes/(?P<routeName>[^/]+)$": AppMeshResponse.dispatch,
    "{0}/v20190125/meshes/(?P<meshName>.*)/virtualNodes/(?P<virtualNodeName>[^/]+)$": AppMeshResponse.dispatch,
    "{0}/v20190125/meshes/(?P<meshName>.*)/virtualNodes$": AppMeshResponse.dispatch,
}



import json

import boto3
import pytest
from botocore.exceptions import ClientError

from moto import mock_aws


@mock_aws
def test_describe_job_execution():
    client = boto3.client("iot", region_name="eu-west-1")
    name = "my-thing"
    job_id = "TestJob"
    # thing
    thing = client.create_thing(thingName=name)
    assert thing["thingName"] == name
    assert "thingArn" in thing

    # job document
    job_document = {"field": "value"}

    job = client.create_job(
        jobId=job_id,
        targets=[thing["thingArn"]],
        document=json.dumps(job_document),
        description="Description",
        presignedUrlConfig={
            "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
            "expiresInSec": 123,
        },
        targetSelection="CONTINUOUS",
        jobExecutionsRolloutConfig={"maximumPerMinute": 10},
    )

    assert job["jobId"] == job_id
    assert "jobArn" in job
    assert "description" in job

    job_execution = client.describe_job_execution(jobId=job_id, thingName=name)
    assert job_execution["execution"]["jobId"] == job_id
    assert job_execution["execution"]["status"] == "QUEUED"
    assert job_execution["execution"]["forceCanceled"] is False
    assert job_execution["execution"]["statusDetails"] == {"detailsMap": {}}
    assert job_execution["execution"]["thingArn"] == thing["thingArn"]
    assert "queuedAt" in job_execution["execution"]
    assert "startedAt" in job_execution["execution"]
    assert "lastUpdatedAt" in job_execution["execution"]
    assert job_execution["execution"]["executionNumber"] == 123
    assert job_execution["execution"]["versionNumber"] == 123
    assert job_execution["execution"]["approximateSecondsBeforeTimedOut"] == 123

    job_execution = client.describe_job_execution(
        jobId=job_id, thingName=name, executionNumber=123
    )
    assert "execution" in job_execution
    assert job_execution["execution"]["jobId"] == job_id
    assert job_execution["execution"]["status"] == "QUEUED"
    assert job_execution["execution"]["forceCanceled"] is False
    assert job_execution["execution"]["statusDetails"] == {"detailsMap": {}}
    assert job_execution["execution"]["thingArn"] == thing["thingArn"]
    assert "queuedAt" in job_execution["execution"]
    assert "startedAt" in job_execution["execution"]
    assert "lastUpdatedAt" in job_execution["execution"]
    assert job_execution["execution"]["executionNumber"] == 123
    assert job_execution["execution"]["versionNumber"] == 123
    assert job_execution["execution"]["approximateSecondsBeforeTimedOut"] == 123

    with pytest.raises(ClientError) as exc:
        client.describe_job_execution(jobId=job_id, thingName=name, executionNumber=456)
    error_code = exc.value.response["Error"]["Code"]
    assert error_code == "ResourceNotFoundException"


@mock_aws
def test_cancel_job_execution():
    client = boto3.client("iot", region_name="eu-west-1")
    name = "my-thing"
    job_id = "TestJob"
    # thing
    thing = client.create_thing(thingName=name)
    assert thing["thingName"] == name
    assert "thingArn" in thing

    # job document
    job_document = {"field": "value"}

    job = client.create_job(
        jobId=job_id,
        targets=[thing["thingArn"]],
        document=json.dumps(job_document),
        description="Description",
        presignedUrlConfig={
            "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
            "expiresInSec": 123,
        },
        targetSelection="CONTINUOUS",
        jobExecutionsRolloutConfig={"maximumPerMinute": 10},
    )

    assert job["jobId"] == job_id
    assert "jobArn" in job
    assert "description" in job

    client.cancel_job_execution(jobId=job_id, thingName=name)
    job_execution = client.describe_job_execution(jobId=job_id, thingName=name)
    assert "execution" in job_execution
    assert job_execution["execution"]["status"] == "CANCELED"


@mock_aws
def test_delete_job_execution():
    client = boto3.client("iot", region_name="eu-west-1")
    name = "my-thing"
    job_id = "TestJob"
    # thing
    thing = client.create_thing(thingName=name)
    assert thing["thingName"] == name
    assert "thingArn" in thing

    # job document
    job_document = {"field": "value"}

    job = client.create_job(
        jobId=job_id,
        targets=[thing["thingArn"]],
        document=json.dumps(job_document),
        description="Description",
        presignedUrlConfig={
            "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
            "expiresInSec": 123,
        },
        targetSelection="CONTINUOUS",
        jobExecutionsRolloutConfig={"maximumPerMinute": 10},
    )

    assert job["jobId"] == job_id
    assert "jobArn" in job
    assert "description" in job

    client.delete_job_execution(jobId=job_id, thingName=name, executionNumber=123)

    with pytest.raises(ClientError) as exc:
        client.describe_job_execution(jobId=job_id, thingName=name, executionNumber=123)
    error_code = exc.value.response["Error"]["Code"]
    assert error_code == "ResourceNotFoundException"


@mock_aws
def test_list_job_executions_for_job():
    client = boto3.client("iot", region_name="eu-west-1")
    name = "my-thing"
    job_id = "TestJob"
    # thing
    thing = client.create_thing(thingName=name)
    assert thing["thingName"] == name
    assert "thingArn" in thing

    # job document
    job_document = {"field": "value"}

    job = client.create_job(
        jobId=job_id,
        targets=[thing["thingArn"]],
        document=json.dumps(job_document),
        description="Description",
        presignedUrlConfig={
            "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
            "expiresInSec": 123,
        },
        targetSelection="CONTINUOUS",
        jobExecutionsRolloutConfig={"maximumPerMinute": 10},
    )

    assert job["jobId"] == job_id
    assert "jobArn" in job
    assert "description" in job

    job_execution = client.list_job_executions_for_job(jobId=job_id)
    assert job_execution["executionSummaries"][0]["thingArn"] == thing["thingArn"]

    job_execution = client.list_job_executions_for_job(jobId=job_id, status="QUEUED")
    assert job_execution["executionSummaries"][0]["thingArn"] == thing["thingArn"]


@mock_aws
def test_list_job_executions_for_thing():
    client = boto3.client("iot", region_name="eu-west-1")
    name = "my-thing"
    job_id = "TestJob"
    # thing
    thing = client.create_thing(thingName=name)
    assert thing["thingName"] == name
    assert "thingArn" in thing

    # job document
    job_document = {"field": "value"}

    job = client.create_job(
        jobId=job_id,
        targets=[thing["thingArn"]],
        document=json.dumps(job_document),
        description="Description",
        presignedUrlConfig={
            "roleArn": "arn:aws:iam::1:role/service-role/iot_job_role",
            "expiresInSec": 123,
        },
        targetSelection="CONTINUOUS",
        jobExecutionsRolloutConfig={"maximumPerMinute": 10},
    )

    assert job["jobId"] == job_id
    assert "jobArn" in job
    assert "description" in job

    job_execution = client.list_job_executions_for_thing(thingName=name)
    assert job_execution["executionSummaries"][0]["jobId"] == job_id

    job_execution = client.list_job_executions_for_thing(
        thingName=name, status="QUEUED"
    )
    assert job_execution["executionSummaries"][0]["jobId"] == job_id


@mock_aws
def test_list_job_executions_for_thing_paginated():
    client = boto3.client("iot", region_name="eu-west-1")
    name = "my-thing"
    thing = client.create_thing(thingName=name)

    for idx in range(0, 10):
        client.create_job(
            jobId=f"TestJob_{idx}",
            targets=[thing["thingArn"]],
            document=json.dumps({"field": "value"}),
        )

    res = client.list_job_executions_for_thing(thingName=name, maxResults=2)
    executions = res["executionSummaries"]
    assert len(executions) == 2

    res = client.list_job_executions_for_thing(
        thingName=name, maxResults=1, nextToken=res["nextToken"]
    )
    executions = res["executionSummaries"]
    assert len(executions) == 1

    res = client.list_job_executions_for_thing(
        thingName=name, nextToken=res["nextToken"]
    )
    executions = res["executionSummaries"]
    assert len(executions) == 7
    assert "nextToken" not in res



"""Handles incoming directconnect requests, invokes methods, returns responses."""

import json

from moto.core.responses import BaseResponse

from .models import Connection, DirectConnectBackend, directconnect_backends


class DirectConnectResponse(BaseResponse):
    """Handler for DirectConnect requests and responses."""

    def __init__(self) -> None:
        super().__init__(service_name="directconnect")

    @property
    def directconnect_backend(self) -> DirectConnectBackend:
        return directconnect_backends[self.current_account][self.region]

    def describe_connections(self) -> str:
        params = json.loads(self.body)
        connections = self.directconnect_backend.describe_connections(
            connection_id=params.get("connectionId"),
        )
        return json.dumps(
            dict(connections=[connection.to_dict() for connection in connections])
        )

    def create_connection(self) -> str:
        params = json.loads(self.body)
        connection: Connection = self.directconnect_backend.create_connection(
            location=params.get("location"),
            bandwidth=params.get("bandwidth"),
            connection_name=params.get("connectionName"),
            lag_id=params.get("lagId"),
            tags=params.get("tags"),
            provider_name=params.get("providerName"),
            request_mac_sec=params.get("requestMACSec"),
        )
        return json.dumps(connection.to_dict())

    def delete_connection(self) -> str:
        params = json.loads(self.body)
        connection: Connection = self.directconnect_backend.delete_connection(
            connection_id=params.get("connectionId"),
        )
        return json.dumps(connection.to_dict())

    def update_connection(self) -> str:
        params = json.loads(self.body)
        connection: Connection = self.directconnect_backend.update_connection(
            connection_id=params.get("connectionId"),
            new_connection_name=params.get("connectionName"),
            new_encryption_mode=params.get("encryptionMode"),
        )
        return json.dumps(connection.to_dict())



from datetime import datetime

import boto3
import pytest
from botocore.client import ClientError

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_get_schedule():
    client = boto3.client("scheduler", region_name="eu-west-1")
    arn = client.create_schedule(
        Name="my-schedule",
        ScheduleExpression="some cron",
        FlexibleTimeWindow={
            "MaximumWindowInMinutes": 4,
            "Mode": "OFF",
        },
        Target={
            "Arn": "not supported yet",
            "RoleArn": "n/a",
        },
        ActionAfterCompletion="DELETE",
    )["ScheduleArn"]

    assert (
        arn
        == f"arn:aws:scheduler:eu-west-1:{DEFAULT_ACCOUNT_ID}:schedule/default/my-schedule"
    )

    resp = client.get_schedule(Name="my-schedule")
    assert resp["Arn"] == arn
    assert resp["Name"] == "my-schedule"
    assert resp["ScheduleExpression"] == "some cron"
    assert resp["FlexibleTimeWindow"] == {
        "MaximumWindowInMinutes": 4,
        "Mode": "OFF",
    }
    assert resp["ActionAfterCompletion"] == "DELETE"
    assert resp["Target"] == {
        "Arn": "not supported yet",
        "RoleArn": "n/a",
        "RetryPolicy": {"MaximumEventAgeInSeconds": 86400, "MaximumRetryAttempts": 185},
    }
    assert isinstance(resp["CreationDate"], datetime)
    assert isinstance(resp["LastModificationDate"], datetime)
    assert resp["CreationDate"] == resp["LastModificationDate"]


@mock_aws
def test_create_get_delete__in_different_group():
    client = boto3.client("scheduler", region_name="eu-west-1")

    client.create_schedule_group(Name="sg")
    schedule_arn = client.create_schedule(
        Name="my-schedule",
        GroupName="sg",
        ScheduleExpression="some cron",
        FlexibleTimeWindow={
            "MaximumWindowInMinutes": 4,
            "Mode": "OFF",
        },
        Target={
            "Arn": "not supported yet",
            "RoleArn": "n/a",
        },
    )["ScheduleArn"]

    assert (
        schedule_arn
        == "arn:aws:scheduler:eu-west-1:123456789012:schedule/sg/my-schedule"
    )

    schedule = client.get_schedule(GroupName="sg", Name="my-schedule")
    assert schedule["Arn"] == schedule_arn

    client.delete_schedule(GroupName="sg", Name="my-schedule")

    with pytest.raises(ClientError) as exc:
        client.get_schedule(GroupName="sg", Name="my-schedule")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@pytest.mark.parametrize(
    "extra_kwargs",
    [
        ({}),
        ({"GroupName": "some-group"}),
    ],
    ids=["without_group", "with_group"],
)
@mock_aws
def test_update_schedule(extra_kwargs):
    client = boto3.client("scheduler", region_name="eu-west-1")

    client.create_schedule_group(Name="some-group")

    client.create_schedule(
        **extra_kwargs,
        Name="my-schedule",
        ScheduleExpression="some cron",
        FlexibleTimeWindow={
            "MaximumWindowInMinutes": 4,
            "Mode": "OFF",
        },
        Target={
            "Arn": "not supported yet",
            "RoleArn": "n/a",
        },
    )

    client.update_schedule(
        **extra_kwargs,
        Name="my-schedule",
        Description="new desc",
        ScheduleExpression="new cron",
        FlexibleTimeWindow={
            "MaximumWindowInMinutes": 4,
            "Mode": "OFF",
        },
        State="DISABLED",
        Target={
            "Arn": "different arn",
            "RoleArn": "n/a",
        },
    )

    schedule = client.get_schedule(**extra_kwargs, Name="my-schedule")
    assert schedule["Description"] == "new desc"
    assert schedule["ScheduleExpression"] == "new cron"
    assert schedule["State"] == "DISABLED"
    assert schedule["Target"] == {
        "Arn": "different arn",
        "RoleArn": "n/a",
        "RetryPolicy": {"MaximumEventAgeInSeconds": 86400, "MaximumRetryAttempts": 185},
    }

    assert isinstance(schedule["CreationDate"], datetime)
    assert isinstance(schedule["LastModificationDate"], datetime)
    assert schedule["CreationDate"] != schedule["LastModificationDate"]


@mock_aws
def test_create_duplicate_schedule():
    client = boto3.client("scheduler", region_name="us-east-1")
    params = {
        "ScheduleExpression": "at(2022-12-12T00:00:00)",
        "FlexibleTimeWindow": {
            "MaximumWindowInMinutes": 4,
            "Mode": "FLEXIBLE",
        },
        "Target": {"Arn": "arn1", "RoleArn": "arn2"},
    }

    client.create_schedule(Name="schedule1", **params)
    with pytest.raises(ClientError) as exc:
        client.create_schedule(Name="schedule1", **params)
    err = exc.value.response["Error"]
    assert err["Code"] == "ConflictException"
    assert err["Message"] == "Schedule schedule1 already exists."


@mock_aws
def test_get_schedule_for_unknown_group():
    client = boto3.client("scheduler", region_name="eu-west-1")

    with pytest.raises(ClientError) as exc:
        client.get_schedule(GroupName="unknown", Name="my-schedule")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"


@mock_aws
def test_get_schedule_for_none_existing_schedule():
    client = boto3.client("scheduler", region_name="eu-west-1")

    with pytest.raises(ClientError) as exc:
        client.get_schedule(Name="my-schedule")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == "Schedule my-schedule does not exist."


@mock_aws
def test_list_schedules():
    client = boto3.client("scheduler", region_name="eu-west-1")

    schedules = client.list_schedules()["Schedules"]
    assert schedules == []

    client.create_schedule_group(Name="group2")

    for group in ["default", "group2"]:
        for schedule in ["sch1", "sch2"]:
            for state in ["ENABLED", "DISABLED"]:
                client.create_schedule(
                    Name=f"{schedule}_{state}",
                    GroupName=group,
                    State=state,
                    ScheduleExpression="some cron",
                    FlexibleTimeWindow={"MaximumWindowInMinutes": 4, "Mode": "OFF"},
                    Target={"Arn": "not supported yet", "RoleArn": "n/a"},
                )

    schedules = client.list_schedules()["Schedules"]
    assert len(schedules) == 8
    # The ListSchedules command should not return the entire Target-dictionary
    assert schedules[0]["Target"] == {"Arn": "not supported yet"}

    schedules = client.list_schedules(GroupName="group2")["Schedules"]
    assert len(schedules) == 4

    schedules = client.list_schedules(State="ENABLED")["Schedules"]
    assert len(schedules) == 4


@mock_aws
def test_delete_schedule_for_none_existing_schedule():
    client = boto3.client("scheduler", region_name="eu-west-1")

    with pytest.raises(ClientError) as exc:
        client.delete_schedule(Name="my-schedule")
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == "Schedule my-schedule does not exist."



import time

import boto3

from moto import mock_aws
from moto.core import DEFAULT_ACCOUNT_ID as ACCOUNT_ID


@mock_aws
def test_initiate_job():
    client = boto3.client("glacier", region_name="us-west-2")
    client.create_vault(vaultName="myname")

    archive = client.upload_archive(vaultName="myname", body=b"body of archive")

    job = client.initiate_job(
        vaultName="myname",
        jobParameters={"ArchiveId": archive["archiveId"], "Type": "archive-retrieval"},
    )
    assert job["ResponseMetadata"]["HTTPStatusCode"] == 202

    headers = job["ResponseMetadata"]["HTTPHeaders"]
    assert "x-amz-job-id" in headers
    # Should be an exact match, but Flask adds 'http' to the start of the Location-header
    assert headers["location"] == "//vaults/myname/jobs/" + headers["x-amz-job-id"]

    # Don't think this is correct - the spec says no body is returned, only headers
    # https://docs.aws.amazon.com/amazonglacier/latest/dev/api-initiate-job-post.html
    assert "jobId" in job
    assert "location" in job


@mock_aws
def test_describe_job():
    client = boto3.client("glacier", region_name="us-west-2")
    client.create_vault(vaultName="myname")

    archive = client.upload_archive(vaultName="myname", body=b"body of archive")

    job = client.initiate_job(
        vaultName="myname",
        jobParameters={"ArchiveId": archive["archiveId"], "Type": "archive-retrieval"},
    )
    job_id = job["jobId"]

    describe = client.describe_job(vaultName="myname", jobId=job_id)
    assert describe["JobId"] == job_id
    assert describe["Action"] == "ArchiveRetrieval"
    assert describe["ArchiveId"] == archive["archiveId"]
    assert (
        describe["VaultARN"] == f"arn:aws:glacier:us-west-2:{ACCOUNT_ID}:vaults/myname"
    )
    assert "CreationDate" in describe
    assert describe["Completed"] is False
    assert describe["StatusCode"] == "InProgress"
    assert describe["ArchiveSizeInBytes"] == 0
    assert describe["InventorySizeInBytes"] == 0
    assert describe["Tier"] == "Standard"


@mock_aws
def test_list_jobs():
    client = boto3.client("glacier", region_name="us-west-2")
    client.create_vault(vaultName="myname")

    archive1 = client.upload_archive(vaultName="myname", body=b"first archive")
    archive2 = client.upload_archive(vaultName="myname", body=b"second archive")

    job1 = client.initiate_job(
        vaultName="myname",
        jobParameters={"ArchiveId": archive1["archiveId"], "Type": "archive-retrieval"},
    )

    job2 = client.initiate_job(
        vaultName="myname",
        jobParameters={"ArchiveId": archive2["archiveId"], "Type": "archive-retrieval"},
    )
    jobs = client.list_jobs(vaultName="myname")["JobList"]

    # Verify the created jobs are in this list
    found_jobs = [j["JobId"] for j in jobs]
    assert job1["jobId"] in found_jobs
    assert job2["jobId"] in found_jobs

    found_job1 = [j for j in jobs if j["JobId"] == job1["jobId"]][0]
    assert found_job1["ArchiveId"] == archive1["archiveId"]
    found_job2 = [j for j in jobs if j["JobId"] == job2["jobId"]][0]
    assert found_job2["ArchiveId"] == archive2["archiveId"]

    # Verify all jobs follow the correct format
    for job in jobs:
        assert "JobId" in job
        assert "Action" in job
        assert "ArchiveId" in job
        assert "VaultARN" in job
        assert "CreationDate" in job
        assert "ArchiveSizeInBytes" in job
        assert "Completed" in job
        assert "StatusCode" in job
        assert "InventorySizeInBytes" in job
        assert "Tier" in job


@mock_aws
def test_get_job_output():
    client = boto3.client("glacier", region_name="us-west-2")
    client.create_vault(vaultName="myname")

    archive = client.upload_archive(vaultName="myname", body=b"contents of archive")

    job = client.initiate_job(
        vaultName="myname",
        jobParameters={"ArchiveId": archive["archiveId"], "Type": "archive-retrieval"},
    )

    output = None
    start = time.time()
    while (time.time() - start) < 10:
        try:
            output = client.get_job_output(vaultName="myname", jobId=job["jobId"])
            break
        except Exception:
            time.sleep(1)

    assert output["status"] == 200
    assert output["contentType"] == "application/octet-stream"
    assert "body" in output

    body = output["body"].read().decode("utf-8")
    assert body == "contents of archive"



"""Unit tests for timestreamquery-supported APIs."""

import boto3
import pytest
import requests
from botocore.exceptions import ClientError

from moto import mock_aws, settings
from tests import DEFAULT_ACCOUNT_ID

# See our Development Tips on writing tests for hints on how to write good tests:
# http://docs.getmoto.org/en/latest/docs/contributing/development_tips/tests.html


@mock_aws
def test_create_scheduled_query():
    target_config = {
        "TimestreamConfiguration": {
            "DatabaseName": "mydb",
            "TableName": "mytab",
            "TimeColumn": "tc",
            "DimensionMappings": [],
        }
    }

    client = boto3.client("timestream-query", region_name="us-east-2")
    arn = client.create_scheduled_query(
        Name="myquery",
        QueryString="SELECT *",
        ScheduleConfiguration={"ScheduleExpression": "* * * * * 1"},
        NotificationConfiguration={"SnsConfiguration": {"TopicArn": "arn:some:topic"}},
        TargetConfiguration=target_config,
        ScheduledQueryExecutionRoleArn="some role",
        ErrorReportConfiguration=get_error_config(),
        KmsKeyId="arn:kms:key",
    )["Arn"]

    assert (
        arn
        == f"arn:aws:timestream:us-east-2:{DEFAULT_ACCOUNT_ID}:scheduled-query/myquery"
    )

    query = client.describe_scheduled_query(ScheduledQueryArn=arn)["ScheduledQuery"]
    assert query["Arn"] == arn
    assert query["Name"] == "myquery"
    assert query["QueryString"] == "SELECT *"
    assert query["CreationTime"]
    assert query["State"] == "ENABLED"
    assert query["ScheduleConfiguration"] == {"ScheduleExpression": "* * * * * 1"}
    assert query["NotificationConfiguration"] == {
        "SnsConfiguration": {"TopicArn": "arn:some:topic"}
    }
    assert query["TargetConfiguration"] == target_config
    assert query["ScheduledQueryExecutionRoleArn"] == "some role"
    assert query["ErrorReportConfiguration"] == get_error_config()
    assert query["KmsKeyId"] == "arn:kms:key"


@mock_aws
def test_delete_scheduled_query():
    client = boto3.client("timestream-query", region_name="us-east-2")

    arn = client.create_scheduled_query(
        Name="myquery",
        QueryString="SELECT *",
        ScheduleConfiguration={"ScheduleExpression": "* * * * * 1"},
        NotificationConfiguration={"SnsConfiguration": {"TopicArn": "arn:some:topic"}},
        ScheduledQueryExecutionRoleArn="some role",
        ErrorReportConfiguration=get_error_config(),
    )["Arn"]
    client.delete_scheduled_query(ScheduledQueryArn=arn)

    with pytest.raises(ClientError) as exc:
        client.describe_scheduled_query(ScheduledQueryArn=arn)
    err = exc.value.response["Error"]
    assert err["Code"] == "ResourceNotFoundException"
    assert err["Message"] == f"The resource with arn {arn} does not exist."


@mock_aws
def test_update_scheduled_query():
    client = boto3.client("timestream-query", region_name="eu-west-1")
    arn = client.create_scheduled_query(
        Name="myquery",
        QueryString="SELECT *",
        ScheduleConfiguration={"ScheduleExpression": "* * * * * 1"},
        NotificationConfiguration={"SnsConfiguration": {"TopicArn": "arn:some:topic"}},
        ScheduledQueryExecutionRoleArn="some role",
        ErrorReportConfiguration=get_error_config(),
    )["Arn"]

    client.update_scheduled_query(
        ScheduledQueryArn=arn,
        State="DISABLED",
    )

    query = client.describe_scheduled_query(ScheduledQueryArn=arn)["ScheduledQuery"]
    assert query["State"] == "DISABLED"


@mock_aws
def test_query_default_results():
    client = boto3.client("timestream-query", region_name="us-east-1")
    resp = client.query(QueryString="SELECT *")

    assert resp["QueryId"]
    assert resp["Rows"] == []
    assert resp["ColumnInfo"] == []


@mock_aws
def test_query__configured_results():
    base_url = (
        "localhost:5000" if settings.TEST_SERVER_MODE else "motoapi.amazonaws.com"
    )

    first_result = {
        "QueryId": "some_id",
        "Rows": [{"Data": [{"ScalarValue": "1"}]}],
        "ColumnInfo": [
            {"Name": "c", "Type": {"ScalarType": "VARCHAR"}},
        ],
        "QueryStatus": {
            "ProgressPercentage": 50,
            "CumulativeBytesScanned": 5,
            "CumulativeBytesMetered": 5,
        },
    }
    second_result = {
        "QueryId": "some_id",
        "Rows": [{"Data": [{"ScalarValue": "1"}]}, {"Data": [{"ScalarValue": "2"}]}],
        "ColumnInfo": [
            {"Name": "c", "Type": {"ScalarType": "VARCHAR"}},
            {"Name": "c", "Type": {"ScalarType": "VARCHAR"}},
        ],
        "QueryStatus": {
            "ProgressPercentage": 100,
            "CumulativeBytesScanned": 10,
            "CumulativeBytesMetered": 10,
        },
    }
    third_result = {
        "Rows": [{"Data": [{"ScalarValue": "5"}]}],
        "ColumnInfo": [{"Name": "c", "Type": {"ScalarType": "VARCHAR"}}],
        "QueryStatus": {
            "ProgressPercentage": 100,
            "CumulativeBytesScanned": 10,
            "CumulativeBytesMetered": 10,
        },
    }

    result = {
        "results": {
            "SELECT *": [first_result, second_result],
            None: [third_result],
        }
    }
    requests.post(
        f"http://{base_url}/moto-api/static/timestream/query-results",
        json=result,
    )

    client = boto3.client("timestream-query", region_name="us-east-1")

    # Unknown QUERY returns third result
    resp = client.query(QueryString="SELECT unknown")

    assert resp["Rows"] == third_result["Rows"]
    assert resp["ColumnInfo"] == third_result["ColumnInfo"]

    # Known QUERY returns first result
    resp = client.query(QueryString="SELECT *")

    assert resp["QueryId"] == "some_id"
    assert resp["Rows"] == first_result["Rows"]
    assert resp["ColumnInfo"] == first_result["ColumnInfo"]

    # Querying the same thing returns the second result
    resp = client.query(QueryString="SELECT *")

    assert resp["QueryId"] == "some_id"
    assert resp["Rows"] == second_result["Rows"]
    assert resp["ColumnInfo"] == second_result["ColumnInfo"]

    # Unknown QUERY returns third result again
    resp = client.query(QueryString="SELECT unknown")

    assert resp["Rows"] == third_result["Rows"]
    assert resp["ColumnInfo"] == third_result["ColumnInfo"]

    # Known QUERY returns nothing - we've already returned all possible results
    resp = client.query(QueryString="SELECT *")

    assert resp["QueryId"]
    assert resp["Rows"] == []
    assert resp["ColumnInfo"] == []


def get_error_config():
    return {"S3Configuration": {"BucketName": "mybucket", "ObjectKeyPrefix": "prefix"}}



# Generated from ASLParser.g4 by ANTLR 4.13.1
from antlr4 import ParseTreeVisitor

if "." in __name__:
    from .ASLParser import ASLParser
else:
    from ASLParser import ASLParser

# This class defines a complete generic visitor for a parse tree produced by ASLParser.


class ASLParserVisitor(ParseTreeVisitor):
    # Visit a parse tree produced by ASLParser#state_machine.
    def visitState_machine(self, ctx: ASLParser.State_machineContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#program_decl.
    def visitProgram_decl(self, ctx: ASLParser.Program_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#top_layer_stmt.
    def visitTop_layer_stmt(self, ctx: ASLParser.Top_layer_stmtContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#startat_decl.
    def visitStartat_decl(self, ctx: ASLParser.Startat_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#comment_decl.
    def visitComment_decl(self, ctx: ASLParser.Comment_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#version_decl.
    def visitVersion_decl(self, ctx: ASLParser.Version_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#state_stmt.
    def visitState_stmt(self, ctx: ASLParser.State_stmtContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#states_decl.
    def visitStates_decl(self, ctx: ASLParser.States_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#state_name.
    def visitState_name(self, ctx: ASLParser.State_nameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#state_decl.
    def visitState_decl(self, ctx: ASLParser.State_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#state_decl_body.
    def visitState_decl_body(self, ctx: ASLParser.State_decl_bodyContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#type_decl.
    def visitType_decl(self, ctx: ASLParser.Type_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#next_decl.
    def visitNext_decl(self, ctx: ASLParser.Next_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#resource_decl.
    def visitResource_decl(self, ctx: ASLParser.Resource_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#input_path_decl.
    def visitInput_path_decl(self, ctx: ASLParser.Input_path_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#result_decl.
    def visitResult_decl(self, ctx: ASLParser.Result_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#result_path_decl.
    def visitResult_path_decl(self, ctx: ASLParser.Result_path_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#output_path_decl.
    def visitOutput_path_decl(self, ctx: ASLParser.Output_path_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#end_decl.
    def visitEnd_decl(self, ctx: ASLParser.End_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#default_decl.
    def visitDefault_decl(self, ctx: ASLParser.Default_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#error_decl.
    def visitError_decl(self, ctx: ASLParser.Error_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#error_path_decl_path.
    def visitError_path_decl_path(self, ctx: ASLParser.Error_path_decl_pathContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#error_path_decl_intrinsic.
    def visitError_path_decl_intrinsic(
        self, ctx: ASLParser.Error_path_decl_intrinsicContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#cause_decl.
    def visitCause_decl(self, ctx: ASLParser.Cause_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#cause_path_decl_path.
    def visitCause_path_decl_path(self, ctx: ASLParser.Cause_path_decl_pathContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#cause_path_decl_intrinsic.
    def visitCause_path_decl_intrinsic(
        self, ctx: ASLParser.Cause_path_decl_intrinsicContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#seconds_decl.
    def visitSeconds_decl(self, ctx: ASLParser.Seconds_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#seconds_path_decl.
    def visitSeconds_path_decl(self, ctx: ASLParser.Seconds_path_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#timestamp_decl.
    def visitTimestamp_decl(self, ctx: ASLParser.Timestamp_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#timestamp_path_decl.
    def visitTimestamp_path_decl(self, ctx: ASLParser.Timestamp_path_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#items_path_decl.
    def visitItems_path_decl(self, ctx: ASLParser.Items_path_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#max_concurrency_decl.
    def visitMax_concurrency_decl(self, ctx: ASLParser.Max_concurrency_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#max_concurrency_path_decl.
    def visitMax_concurrency_path_decl(
        self, ctx: ASLParser.Max_concurrency_path_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#parameters_decl.
    def visitParameters_decl(self, ctx: ASLParser.Parameters_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#timeout_seconds_decl.
    def visitTimeout_seconds_decl(self, ctx: ASLParser.Timeout_seconds_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#timeout_seconds_path_decl.
    def visitTimeout_seconds_path_decl(
        self, ctx: ASLParser.Timeout_seconds_path_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#heartbeat_seconds_decl.
    def visitHeartbeat_seconds_decl(self, ctx: ASLParser.Heartbeat_seconds_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#heartbeat_seconds_path_decl.
    def visitHeartbeat_seconds_path_decl(
        self, ctx: ASLParser.Heartbeat_seconds_path_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_tmpl_decl.
    def visitPayload_tmpl_decl(self, ctx: ASLParser.Payload_tmpl_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_binding_path.
    def visitPayload_binding_path(self, ctx: ASLParser.Payload_binding_pathContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_binding_path_context_obj.
    def visitPayload_binding_path_context_obj(
        self, ctx: ASLParser.Payload_binding_path_context_objContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_binding_intrinsic_func.
    def visitPayload_binding_intrinsic_func(
        self, ctx: ASLParser.Payload_binding_intrinsic_funcContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_binding_value.
    def visitPayload_binding_value(self, ctx: ASLParser.Payload_binding_valueContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#intrinsic_func.
    def visitIntrinsic_func(self, ctx: ASLParser.Intrinsic_funcContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_arr_decl.
    def visitPayload_arr_decl(self, ctx: ASLParser.Payload_arr_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_value_decl.
    def visitPayload_value_decl(self, ctx: ASLParser.Payload_value_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_value_float.
    def visitPayload_value_float(self, ctx: ASLParser.Payload_value_floatContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_value_int.
    def visitPayload_value_int(self, ctx: ASLParser.Payload_value_intContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_value_bool.
    def visitPayload_value_bool(self, ctx: ASLParser.Payload_value_boolContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_value_null.
    def visitPayload_value_null(self, ctx: ASLParser.Payload_value_nullContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#payload_value_str.
    def visitPayload_value_str(self, ctx: ASLParser.Payload_value_strContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#result_selector_decl.
    def visitResult_selector_decl(self, ctx: ASLParser.Result_selector_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#state_type.
    def visitState_type(self, ctx: ASLParser.State_typeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#choices_decl.
    def visitChoices_decl(self, ctx: ASLParser.Choices_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#choice_rule_comparison_variable.
    def visitChoice_rule_comparison_variable(
        self, ctx: ASLParser.Choice_rule_comparison_variableContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#choice_rule_comparison_composite.
    def visitChoice_rule_comparison_composite(
        self, ctx: ASLParser.Choice_rule_comparison_compositeContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#comparison_variable_stmt.
    def visitComparison_variable_stmt(
        self, ctx: ASLParser.Comparison_variable_stmtContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#comparison_composite_stmt.
    def visitComparison_composite_stmt(
        self, ctx: ASLParser.Comparison_composite_stmtContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#comparison_composite.
    def visitComparison_composite(self, ctx: ASLParser.Comparison_compositeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#variable_decl.
    def visitVariable_decl(self, ctx: ASLParser.Variable_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#comparison_func.
    def visitComparison_func(self, ctx: ASLParser.Comparison_funcContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#branches_decl.
    def visitBranches_decl(self, ctx: ASLParser.Branches_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#item_processor_decl.
    def visitItem_processor_decl(self, ctx: ASLParser.Item_processor_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#item_processor_item.
    def visitItem_processor_item(self, ctx: ASLParser.Item_processor_itemContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#processor_config_decl.
    def visitProcessor_config_decl(self, ctx: ASLParser.Processor_config_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#processor_config_field.
    def visitProcessor_config_field(self, ctx: ASLParser.Processor_config_fieldContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#mode_decl.
    def visitMode_decl(self, ctx: ASLParser.Mode_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#mode_type.
    def visitMode_type(self, ctx: ASLParser.Mode_typeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#execution_decl.
    def visitExecution_decl(self, ctx: ASLParser.Execution_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#execution_type.
    def visitExecution_type(self, ctx: ASLParser.Execution_typeContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#iterator_decl.
    def visitIterator_decl(self, ctx: ASLParser.Iterator_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#iterator_decl_item.
    def visitIterator_decl_item(self, ctx: ASLParser.Iterator_decl_itemContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#item_selector_decl.
    def visitItem_selector_decl(self, ctx: ASLParser.Item_selector_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#item_reader_decl.
    def visitItem_reader_decl(self, ctx: ASLParser.Item_reader_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#items_reader_field.
    def visitItems_reader_field(self, ctx: ASLParser.Items_reader_fieldContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#reader_config_decl.
    def visitReader_config_decl(self, ctx: ASLParser.Reader_config_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#reader_config_field.
    def visitReader_config_field(self, ctx: ASLParser.Reader_config_fieldContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#input_type_decl.
    def visitInput_type_decl(self, ctx: ASLParser.Input_type_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#csv_header_location_decl.
    def visitCsv_header_location_decl(
        self, ctx: ASLParser.Csv_header_location_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#csv_headers_decl.
    def visitCsv_headers_decl(self, ctx: ASLParser.Csv_headers_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#max_items_decl.
    def visitMax_items_decl(self, ctx: ASLParser.Max_items_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#max_items_path_decl.
    def visitMax_items_path_decl(self, ctx: ASLParser.Max_items_path_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#tolerated_failure_count_decl.
    def visitTolerated_failure_count_decl(
        self, ctx: ASLParser.Tolerated_failure_count_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#tolerated_failure_count_path_decl.
    def visitTolerated_failure_count_path_decl(
        self, ctx: ASLParser.Tolerated_failure_count_path_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#tolerated_failure_percentage_decl.
    def visitTolerated_failure_percentage_decl(
        self, ctx: ASLParser.Tolerated_failure_percentage_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#tolerated_failure_percentage_path_decl.
    def visitTolerated_failure_percentage_path_decl(
        self, ctx: ASLParser.Tolerated_failure_percentage_path_declContext
    ):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#label_decl.
    def visitLabel_decl(self, ctx: ASLParser.Label_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#retry_decl.
    def visitRetry_decl(self, ctx: ASLParser.Retry_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#retrier_decl.
    def visitRetrier_decl(self, ctx: ASLParser.Retrier_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#retrier_stmt.
    def visitRetrier_stmt(self, ctx: ASLParser.Retrier_stmtContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#error_equals_decl.
    def visitError_equals_decl(self, ctx: ASLParser.Error_equals_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#interval_seconds_decl.
    def visitInterval_seconds_decl(self, ctx: ASLParser.Interval_seconds_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#max_attempts_decl.
    def visitMax_attempts_decl(self, ctx: ASLParser.Max_attempts_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#backoff_rate_decl.
    def visitBackoff_rate_decl(self, ctx: ASLParser.Backoff_rate_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#max_delay_seconds_decl.
    def visitMax_delay_seconds_decl(self, ctx: ASLParser.Max_delay_seconds_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#jitter_strategy_decl.
    def visitJitter_strategy_decl(self, ctx: ASLParser.Jitter_strategy_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#catch_decl.
    def visitCatch_decl(self, ctx: ASLParser.Catch_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#catcher_decl.
    def visitCatcher_decl(self, ctx: ASLParser.Catcher_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#catcher_stmt.
    def visitCatcher_stmt(self, ctx: ASLParser.Catcher_stmtContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#comparison_op.
    def visitComparison_op(self, ctx: ASLParser.Comparison_opContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#choice_operator.
    def visitChoice_operator(self, ctx: ASLParser.Choice_operatorContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#states_error_name.
    def visitStates_error_name(self, ctx: ASLParser.States_error_nameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#error_name.
    def visitError_name(self, ctx: ASLParser.Error_nameContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#json_obj_decl.
    def visitJson_obj_decl(self, ctx: ASLParser.Json_obj_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#json_binding.
    def visitJson_binding(self, ctx: ASLParser.Json_bindingContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#json_arr_decl.
    def visitJson_arr_decl(self, ctx: ASLParser.Json_arr_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#json_value_decl.
    def visitJson_value_decl(self, ctx: ASLParser.Json_value_declContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ASLParser#keyword_or_string.
    def visitKeyword_or_string(self, ctx: ASLParser.Keyword_or_stringContext):
        return self.visitChildren(ctx)


del ASLParser
